import logging
from abc import ABC, abstractmethod
from typing import Optional

from platypush.plugins import Plugin


class MediaSearcher(ABC):
    """
    Base class for media searchers
    """

    def __init__(self, *_, media_plugin: Optional[Plugin] = None, **__):
        from .. import MediaPlugin

        self.logger = logging.getLogger(self.__class__.__name__)
        assert isinstance(
            media_plugin, MediaPlugin
        ), f'Invalid media plugin: {media_plugin}'
        self.media_plugin: Optional[MediaPlugin] = media_plugin

    @abstractmethod
    def search(self, query, *args, **kwargs):
        raise NotImplementedError(
            'The search method should be implemented by a derived class'
        )

    @abstractmethod
    def supports(self, type: str) -> bool:
        raise NotImplementedError(
            'The type method should be implemented by a derived class'
        )


from .local import LocalMediaSearcher  # noqa: E402
from .youtube import YoutubeMediaSearcher  # noqa: E402
from .torrent import TorrentMediaSearcher  # noqa: E402
from .plex import PlexMediaSearcher  # noqa: E402
from .jellyfin import JellyfinMediaSearcher  # noqa: E402

searchers = [
    LocalMediaSearcher,
    YoutubeMediaSearcher,
    TorrentMediaSearcher,
    PlexMediaSearcher,
    JellyfinMediaSearcher,
]

__all__ = [
    'JellyfinMediaSearcher',
    'LocalMediaSearcher',
    'MediaSearcher',
    'PlexMediaSearcher',
    'TorrentMediaSearcher',
    'YoutubeMediaSearcher',
]


# vim:sw=4:ts=4:et:



from dataclasses import dataclass


@dataclass
class MemoryStats:
    """
    Memory stats data class.
    """

    total: int
    available: int
    used: int
    free: int
    active: int
    inactive: int
    buffers: int
    cached: int
    shared: int
    percent: float


@dataclass
class SwapStats:
    """
    Swap memory stats data class.
    """

    total: int
    used: int
    free: int
    percent: float



from dataclasses import dataclass
from typing import Optional


@dataclass
class Fan:
    """
    System fan sensor data class.
    """

    id: str
    label: Optional[str] = None
    value: Optional[float] = None



from enum import Enum
from socket import AddressFamily

from marshmallow import fields, pre_load
from marshmallow.validate import OneOf

from .._base import SystemBaseSchema


class NetworkInterfaceAddressSchema(SystemBaseSchema):
    """
    Schema for network interface address.
    """

    family = fields.String(
        allow_none=True,
        metadata={
            'description': 'The address family.',
            'example': 'AF_INET',
        },
    )

    address = fields.String(
        allow_none=True,
        metadata={
            'description': 'The IP address associated with the interface.',
            'example': '192.168.1.4',
        },
    )

    netmask = fields.String(
        allow_none=True,
        metadata={
            'description': 'The netmask associated with the interface.',
            'example': '255.255.255.0',
        },
    )

    broadcast = fields.String(
        allow_none=True,
        metadata={
            'description': 'The broadcast address associated with the interface.',
            'example': '192.168.1.255',
        },
    )


class NetworkInterfaceSchema(SystemBaseSchema):
    """
    Schema for network interface stats.
    """

    interface = fields.String(
        allow_none=True,
        metadata={
            'description': 'The name of the network interface.',
            'example': 'eth0',
        },
    )

    bytes_sent = fields.Integer(
        missing=0,
        metadata={
            'description': 'The number of bytes sent.',
            'example': 123456,
        },
    )

    bytes_recv = fields.Integer(
        missing=0,
        metadata={
            'description': 'The number of bytes received.',
            'example': 654321,
        },
    )

    packets_sent = fields.Integer(
        missing=0,
        metadata={
            'description': 'The number of packets sent.',
            'example': 123,
        },
    )

    packets_recv = fields.Integer(
        missing=0,
        metadata={
            'description': 'The number of packets received.',
            'example': 321,
        },
    )

    errors_in = fields.Integer(
        missing=0,
        metadata={
            'description': 'The number of errors on the input side.',
            'example': 10,
        },
    )

    errors_out = fields.Integer(
        missing=0,
        metadata={
            'description': 'The number of errors on the output side.',
            'example': 5,
        },
    )

    drop_in = fields.Integer(
        missing=0,
        metadata={
            'description': 'The number of dropped packets on the input side.',
            'example': 1,
        },
    )

    drop_out = fields.Integer(
        missing=0,
        metadata={
            'description': 'The number of dropped packets on the output side.',
            'example': 2,
        },
    )

    is_up = fields.Boolean(
        metadata={
            'description': 'Whether the interface is up.',
            'example': True,
        },
    )

    speed = fields.Integer(
        metadata={
            'description': 'The advertised speed of the interface in Mbps.',
            'example': 1000,
        },
    )

    mtu = fields.Integer(
        metadata={
            'description': 'The maximum transmission unit of the interface in bytes.',
            'example': 1500,
        },
    )

    duplex = fields.String(
        validate=OneOf(['FULL', 'HALF', 'UNKNOWN']),
        metadata={
            'description': 'Interface duplex configuration. Can be FULL, '
            'HALF or UNKNOWN',
            'example': 'FULL',
        },
    )

    flags = fields.List(
        fields.String(),
        metadata={
            'description': 'A list of flags associated with the interface.',
            'example': ['up', 'broadcast', 'running', 'multicast'],
        },
    )

    addresses = fields.List(
        fields.Nested(NetworkInterfaceAddressSchema),
        metadata={
            'description': 'A list of addresses associated with the interface.',
            'example': [
                {
                    'family': 'AF_INET',
                    'address': '192.168.1.4',
                    'netmask': '255.255.255.0',
                    'broadcast': '192.168.1.255',
                },
            ],
        },
    )

    @pre_load
    def pre_load(self, data: dict, **_) -> dict:
        data = super().pre_load(data)

        # Custom attribute mappings
        for in_attr, out_attr in {
            'errin': 'errors_in',
            'errout': 'errors_out',
            'dropin': 'drop_in',
            'dropout': 'drop_out',
            'isup': 'is_up',
        }.items():
            if in_attr in data:
                data[out_attr] = data.pop(in_attr)

        # Serialize enum values
        for i, addr in enumerate(data.get('addresses', [])):
            if hasattr(addr, '_asdict'):
                addr = addr._asdict()
            if isinstance(addr.get('family'), AddressFamily):
                addr['family'] = addr['family'].name
            data['addresses'][i] = addr

        if isinstance(data.get('duplex'), Enum):
            data['duplex'] = data['duplex'].name.split('_')[-1]

        # Split the flags string
        data['flags'] = data.get('flags', '').split(',')

        return data



import inspect
import json
import re
import subprocess
from dataclasses import fields
from typing import Dict, Optional

from .. import YoutubeMediaResource
from ._base import MediaResourceParser


class YoutubeResourceParser(MediaResourceParser):
    """
    Parser for yt-dlp-compatible resources.
    """

    @staticmethod
    def _get_extractors():
        try:
            from yt_dlp.extractor import _extractors  # type: ignore
        except ImportError:
            # yt-dlp not installed
            return

        for _, obj_type in inspect.getmembers(_extractors):
            if (
                inspect.isclass(obj_type)
                and isinstance(getattr(obj_type, "_VALID_URL", None), str)
                and obj_type.__name__ != "GenericIE"
            ):
                yield obj_type

    @classmethod
    def is_youtube_resource(cls, resource: str):
        return any(
            re.search(getattr(extractor, '_VALID_URL', '^$'), resource)
            for extractor in cls._get_extractors()
        )

    @staticmethod
    def _get_youtube_best_thumbnail(info: Dict[str, dict]):
        thumbnails = info.get('thumbnails', {})
        if not thumbnails:
            return None

        # Preferred resolution
        for res in ((640, 480), (480, 360), (320, 240)):
            thumb = next(
                (
                    thumb
                    for thumb in thumbnails
                    if thumb.get('width') == res[0] and thumb.get('height') == res[1]
                ),
                None,
            )

            if thumb:
                return thumb.get('url')

        # Default fallback (best quality)
        return info.get('thumbnail')

    def parse(
        self,
        resource: str,
        *_,
        youtube_format: Optional[str] = None,
        youtube_audio_format: Optional[str] = None,
        only_audio: bool = False,
        **__
    ) -> Optional[YoutubeMediaResource]:
        if not self.is_youtube_resource(resource):
            return None

        youtube_format = youtube_format or self._media.youtube_format
        if only_audio:
            youtube_format = (
                youtube_audio_format
                or self._media.youtube_audio_format
                or youtube_format
            )

        ytdl_cmd = [
            self._media._ytdl,  # pylint: disable=protected-access
            *(['-f', youtube_format] if youtube_format else []),
            *(['-x'] if only_audio else []),
            '-j',
            '-g',
            resource,
        ]

        self.logger.info('Executing command %s', ' '.join(ytdl_cmd))
        with subprocess.Popen(ytdl_cmd, stdout=subprocess.PIPE) as ytdl:
            output = ytdl.communicate()[0].decode().strip()
            ytdl.wait()

        self.logger.debug('yt-dlp output: %s', output)
        lines = output.split('\n')

        if not lines:
            self.logger.warning('No output from yt-dlp')
            return None

        info = json.loads(lines[-1])
        args = {
            **{
                field.name: info.get(field.name)
                for field in fields(YoutubeMediaResource)
            },
            'url': info.get('webpage_url'),
            'image': self._get_youtube_best_thumbnail(info),
            'type': info.get('extractor'),
            'media_plugin': self._media,
        }

        return YoutubeMediaResource(**args)


# vim:sw=4:ts=4:et:



from typing import List, Dict, Any, Optional, Tuple, Union

from platypush.plugins import Plugin, action
from platypush.user import UserManager


class UserPlugin(Plugin):
    """
    Plugin to programmatically create and manage users and user sessions
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.user_manager = UserManager()

    @action
    def create_user(
        self,
        username,
        password,
        executing_user=None,
        executing_user_password=None,
        session_token=None,
        **kwargs,
    ):
        """
        Create a user. This action needs to be executed by an already existing
        user, who needs to authenticate with their own credentials, unless this
        is the first user created on the system.

        :return:

            .. code-block:: json

                {
                    "user_id": 1,
                    "username": "test-user",
                    "created_at": "2020-11-26T22:41:40.550574"
                }

        """

        if (
            self.user_manager.get_user_count() > 0
            and not executing_user
            and not session_token
        ):
            return None, "You need to authenticate in order to create another user"

        if not self.user_manager.authenticate_user(
            executing_user, executing_user_password
        ):
            user, _ = self.user_manager.authenticate_user_session(session_token)[:2]
            if not user:
                return None, "Invalid credentials and/or session_token"

        try:
            user = self.user_manager.create_user(username, password, **kwargs)
        except (NameError, ValueError) as e:
            return None, str(e)

        return {
            'user_id': user.user_id,
            'username': user.username,
            'created_at': user.created_at.isoformat(),
        }

    @action
    def authenticate_user(
        self,
        username: str,
        password: str,
        code: Optional[str] = None,
        return_details: bool = False,
    ) -> Union[bool, Tuple[bool, str]]:
        """
        Authenticate a user.

        :param username: Username.
        :param password: Password.
        :param code: Optional 2FA code, if 2FA is enabled for the user.
        :param return_details: If True then return the error details in
            case of authentication failure.
        :return: If ``return_details`` is False (default), the action returns
            True if the provided credentials are valid, False otherwise.
            If ``return_details`` is True then the action returns a tuple
            (authenticated, error_details) where ``authenticated`` is True if
            the provided credentials are valid, False otherwise, and
            ``error_details`` is a string containing the error details in case
            of authentication failure. Supported error details are:

                - ``invalid_credentials``: Invalid username or password.
                - ``invalid_otp_code``: Invalid 2FA code.
                - ``missing_otp_code``: Username/password are correct, but a 2FA
                  code is required for the user.

        """
        response = self.user_manager.authenticate_user(
            username, password, code=code, with_status=return_details
        )

        if return_details:
            assert (
                isinstance(response, tuple) and len(response) == 2
            ), 'Invalid response from authenticate_user'
            return response[0], response[1].value

        return response

    @action
    def update_password(self, username, old_password, new_password):
        """
        Update the password of a user.

        :return: True if the password was successfully updated, false otherwise
        """

        return self.user_manager.update_password(username, old_password, new_password)

    @action
    def delete_user(
        self,
        username,
        executing_user=None,
        executing_user_password=None,
        session_token=None,
    ):
        """
        Delete a user.
        """

        if not self.user_manager.authenticate_user(
            executing_user, executing_user_password
        ):
            user, _ = self.user_manager.authenticate_user_session(session_token)[:2]
            if not user:
                return None, "Invalid credentials and/or session_token"

        try:
            return self.user_manager.delete_user(username)
        except NameError:
            return None, f"No such user: {username}"

    @action
    def create_session(self, username, password, code=None, expires_at=None):
        """
        Create a user session.

        :return:

            .. code-block:: json

                {
                    "session_token": "secret",
                    "user_id": 1,
                    "username": "test-user",
                    "created_at": "2020-11-26T22:41:40.550574",
                    "expires_at": "2020-11-26T22:41:40.550574"
                }

        """

        session = self.user_manager.create_user_session(
            username=username, password=password, code=code, expires_at=expires_at
        )

        if isinstance(session, tuple):
            session = session[0]

        if not session:
            return None, "Invalid credentials"

        return {
            'session_token': session.session_token,
            'user_id': session.user_id,
            'created_at': session.created_at.isoformat(),
            'expires_at': (
                session.expires_at.isoformat() if session.expires_at else None  # type: ignore
            ),
        }

    @action
    def authenticate_session(self, session_token):
        """
        Authenticate a session by token and return the associated user.

        :return:

            .. code-block:: json

                {
                    "user_id": 1,
                    "username": "test-user",
                    "created_at": "2020-11-26T22:41:40.550574"
                }

        """

        user, _ = self.user_manager.authenticate_user_session(session_token)[:2]
        if not user:
            return None, 'Invalid session token'

        return {
            'user_id': user.user_id,
            'username': user.username,
            'created_at': user.created_at.isoformat(),
        }

    @action
    def delete_session(self, session_token):
        """
        Delete a user session.
        """

        return self.user_manager.delete_user_session(session_token)

    @action
    def get_users(self) -> List[Dict[str, Any]]:
        """
        Get the list of registered users.

        :return:

            .. code-block:: json

                [
                    {
                        "user_id": 1,
                        "username": "user1",
                        "created_at": "2020-11-26T22:41:40.550574"
                    },
                    {
                        "user_id": 2,
                        "username": "user2",
                        "created_at": "2020-11-28T21:10:23.224813"
                    }
                ]

        """
        users = self.user_manager.get_users()
        return [
            {
                'user_id': user.user_id,
                'username': user.username,
                'created_at': user.created_at.isoformat(),
            }
            for user in users
        ]

    @action
    def get_user_by_session(self, session_token: str) -> dict:
        """
        Get the user record associated to a session token.

        :param session_token: Session token.
        :return:

            .. code-block:: json

                [
                    {
                        "user_id": 1,
                        "username": "user1",
                        "created_at": "2020-11-26T22:41:40.550574"
                    }
                ]

        """
        user = self.user_manager.get_user_by_session(session_token)
        assert user, 'No user associated with the specified session token'

        return {
            'user_id': user.user_id,
            'username': user.username,
            'created_at': user.created_at.isoformat(),
        }


# vim:sw=4:ts=4:et:



from dataclasses import dataclass
from typing import Optional


@dataclass
class Disk:
    """
    Disk data class.
    """

    device: str
    mountpoint: Optional[str] = None
    fstype: Optional[str] = None
    opts: Optional[str] = None
    total: Optional[int] = None
    used: Optional[int] = None
    free: Optional[int] = None
    read_count: Optional[int] = None
    write_count: Optional[int] = None
    read_bytes: Optional[int] = None
    write_bytes: Optional[int] = None
    read_time: Optional[float] = None
    write_time: Optional[float] = None
    busy_time: Optional[float] = None
    percent: Optional[float] = None



from typing import Type

from platypush.common.gstreamer import Pipeline
from platypush.context import get_bus
from platypush.message.event.media import (
    MediaEvent,
    MediaPlayEvent,
    MediaPauseEvent,
    MediaStopEvent,
    NewPlayingMediaEvent,
    MediaMuteChangedEvent,
    MediaSeekEvent,
)
from platypush.plugins.media import MediaResource


class MediaPipeline(Pipeline):
    def __init__(self, resource: MediaResource):
        super().__init__()

        self.resource = resource
        self._first_play = True
        if resource.resource and resource.fd is None:
            self.add_source('playbin', uri=resource.resource)
        elif resource.fd is not None:
            self.add_source('playbin', uri=f'fd://{resource.fd.fileno()}')
        else:
            raise AssertionError('No resource specified')

    def post_event(self, evt_class: Type[MediaEvent], **kwargs):
        kwargs['player'] = 'local'
        kwargs['plugin'] = 'media.gstreamer'

        if self.resource:
            resource_args = self.resource.to_dict()
            resource_args.pop('type', None)
            kwargs.update(resource_args)

        evt = evt_class(**kwargs)
        get_bus().post(evt)

    def on_play(self):
        super().on_play()
        if self._first_play:
            self.post_event(NewPlayingMediaEvent, resource=self.resource)
            self._first_play = False

        self.post_event(MediaPlayEvent)

    def on_pause(self):
        super().on_pause()
        self.post_event(MediaPauseEvent)

    def play(self):
        from gi.repository import Gst

        self._first_play = self.get_state() == Gst.State.NULL
        super().play()

    def stop(self):
        super().stop()
        self._first_play = True
        self.post_event(MediaStopEvent)

    def mute(self):
        super().mute()
        self.post_event(MediaMuteChangedEvent, mute=self.is_muted())

    def unmute(self):
        super().unmute()
        self.post_event(MediaMuteChangedEvent, mute=self.is_muted())

    def seek(self, position: float):
        from gi.repository import Gst

        if not self.source:
            self.logger.info('Cannot seek on a pipeline without a source')
            return

        position = max(0, position)
        duration = self.get_duration()
        if duration and position > duration:
            position = duration

        cur_pos = self.get_position() or 0
        seek_ns = int((position - cur_pos) * 1e9)
        self.source.seek_simple(Gst.Format.TIME, Gst.SeekFlags.FLUSH, seek_ns)
        self.post_event(MediaSeekEvent, position=self.get_position())


# vim:sw=4:ts=4:et:



from typing import Any, Optional

from marshmallow import fields, pre_load, post_dump, Schema, EXCLUDE
from marshmallow.validate import Range

from platypush.schemas import DateTime, EnumField

from ._model import (
    LinodeInstanceBackupScheduleDay,
    LinodeInstanceBackupScheduleWindow,
    LinodeInstanceStatus,
)


class FieldWithId(fields.Function):
    """
    Field that handles values that are objects with an ``id`` attribute.
    """

    def _deserialize(self, value: Any, *_, **__) -> Optional[Any]:
        return value.id if value is not None else None

    def _serialize(self, value: Any, *_, **__) -> Optional[Any]:
        return value


class LinodeBaseSchema(Schema):
    """
    Base schema for Linode objects.
    """

    class Meta:  # type: ignore
        """
        Meta class for the schema.
        """

        unknown = EXCLUDE


class LinodeInstanceSpecsSchema(LinodeBaseSchema):
    """
    Linode instance specifications schema.
    """

    disk = fields.Int(
        metadata={
            'description': 'Allocated disk size, in MB',
            'example': 100000,
        }
    )

    memory = fields.Int(
        metadata={
            'description': 'Allocated RAM size, in MB',
            'example': 8192,
        }
    )

    cpus = fields.Int(
        data_key='vcpus',
        metadata={
            'description': 'Number of virtual CPUs allocated to the instance',
            'example': 4,
        },
    )

    gpus = fields.Int(
        metadata={
            'metadata': {
                'description': 'Number of GPUs allocated to the instance',
                'example': 1,
            }
        }
    )

    transfer = fields.Int(
        metadata={
            'description': (
                'Number of network transfers this instance is allotted each month',
            ),
            'example': 5000,
        }
    )

    @post_dump
    def post_dump(self, data: dict, **_) -> dict:
        if 'vcpus' in data:
            data['cpus'] = data.pop('vcpus')
        return data


class LinodeInstanceAlertsSchema(LinodeBaseSchema):
    """
    Schema that models the alerts configuration of a Linode instance.
    """

    cpu = fields.Int(
        allow_none=True,
        metadata={
            'description': (
                'The percentage of CPU average usage over the past two hours '
                'required to trigger an alert'
            ),
            'example': 90,
        },
    )

    io = fields.Int(
        allow_none=True,
        metadata={
            'description': (
                'The amount of disk I/O operations per second required to '
                'trigger an alert'
            ),
            'example': 5000,
        },
    )

    network_in = fields.Float(
        allow_none=True,
        metadata={
            'description': (
                'The amount of incoming network traffic, in Mbit/s, '
                'required to trigger an alert'
            ),
            'example': 10,
        },
    )

    network_out = fields.Float(
        allow_none=True,
        metadata={
            'description': (
                'The amount of outgoing network traffic, in Mbit/s, '
                'required to trigger an alert'
            ),
            'example': 10,
        },
    )

    transfer_quota = fields.Float(
        allow_none=True,
        metadata={
            'validate': Range(min=0, max=100),
            'description': (
                'The percentage of network transfer that may be used before '
                'an alert is triggered, between 0 and 100'
            ),
            'example': 80,
        },
    )


class LinodeInstanceBackupScheduleSchema(LinodeBaseSchema):
    """
    Schema that models the backup schedule of a Linode instance.
    """

    day = EnumField(
        type=LinodeInstanceBackupScheduleDay,
        allow_none=True,
        metadata={
            'description': 'Day of the week when the backups are scheduled',
            'example': 'Sunday',
        },
    )

    window = EnumField(
        type=LinodeInstanceBackupScheduleWindow,
        allow_none=True,
        metadata={
            'description': 'Time window when the backups are scheduled',
            'example': 'W10',
        },
    )


class LinodeInstanceBackupsSchema(LinodeBaseSchema):
    """
    Schema that models the backup status of a Linode instance.
    """

    available = fields.Bool(
        metadata={
            'description': 'Whether the backups are available for this instance',
            'example': True,
        }
    )

    enabled = fields.Bool(
        metadata={
            'description': 'Whether the backups are enabled on this instance',
            'example': True,
        }
    )

    schedule = fields.Nested(
        LinodeInstanceBackupScheduleSchema,
        metadata={
            'description': 'Backup schedule configuration',
        },
    )

    last_successful = DateTime(
        allow_none=True,
        metadata={
            'description': 'When the last backup was successful',
            'example': '2020-01-01T00:00:00Z',
        },
    )


class LinodeInstanceSchema(LinodeBaseSchema):
    """
    Class that models a Linode instance.
    """

    id = fields.Int(
        required=True,
        metadata={
            'description': 'Instance ID',
            'example': 12345,
        },
    )

    name = fields.String(
        required=True,
        data_key='label',
        metadata={
            'description': 'Instance name',
            'example': 'my-instance',
        },
    )

    instance_type = FieldWithId(
        metadata={
            'description': 'Instance type',
            'example': 'g6-standard-4',
        }
    )

    ipv4_addresses = fields.List(
        fields.String(),
        data_key='ipv4',
        metadata={
            'description': 'List of IPv4 addresses associated with this instance',
            'example': '["1.2.3.4"]',
        },
    )

    ipv6_address = fields.String(
        data_key='ipv6',
        metadata={
            'description': 'IPv6 address associated with this instance',
            'example': '1234:5678::9abc:def0:1234:5678/128',
        },
    )

    group = fields.String(
        metadata={
            'description': 'Group the instance belongs to',
            'example': 'my-group',
        }
    )

    status = EnumField(
        type=LinodeInstanceStatus,
        metadata={
            'description': 'Instance status',
            'example': 'running',
        },
    )

    tags = fields.List(
        fields.String(),
        metadata={
            'description': 'List of tags associated with this instance',
            'example': '["tag1", "tag2"]',
        },
    )

    image = FieldWithId(
        metadata={
            'description': 'Image used to ',
            'example': 'linode/archlinux2014.04',
        }
    )

    region = FieldWithId(
        metadata={
            'description': 'Region where the instance is located',
            'example': 'eu-west',
        }
    )

    hypervisor = fields.String(
        metadata={
            'description': 'The virtualization engine powering this instance',
            'example': 'kvm',
        }
    )

    specs = fields.Nested(LinodeInstanceSpecsSchema)
    alerts = fields.Nested(LinodeInstanceAlertsSchema)
    backups = fields.Nested(LinodeInstanceBackupsSchema)
    created_at = DateTime(
        data_key='created',
        metadata={
            'description': 'Instance creation date',
            'example': '2020-01-01T00:00:00Z',
        },
    )

    updated_at = DateTime(
        data_key='updated',
        metadata={
            'description': 'When the instance was last polled/updated',
            'example': '2020-01-01T01:00:00Z',
        },
    )

    @pre_load
    def pre_load(self, data: dict, **_) -> dict:
        from linode_api4.objects.base import MappedObject

        # Expand MappedObjects to dictionaries
        for key, value in data.items():
            if isinstance(value, MappedObject):
                data[key] = value.dict

        # NOTE Workaround for type -> instance_type not being correctly mapped
        if 'type' in data:
            data['instance_type'] = data.pop('type')

        return data

    @post_dump
    def post_dump(self, data: dict, **_) -> dict:
        for data_key, dump_key in [
            ('label', 'name'),
            ('ipv4', 'ipv4_addresses'),
            ('ipv6', 'ipv6_address'),
            ('created', 'created_at'),
            ('updated', 'updated_at'),
        ]:
            if data_key in data:
                data[dump_key] = data.pop(data_key)

        return data



from datetime import datetime

from dateutil.tz import gettz
from marshmallow import fields, pre_load

from platypush.schemas import DateTime
from platypush.schemas.dataclasses import percent_field
from .._base import SystemBaseSchema


class ProcessSchema(SystemBaseSchema):
    """
    Schema for system processes.
    """

    pid = fields.Int(
        missing=-1,
        metadata={
            'description': 'The process ID.',
            'example': 1234,
        },
    )

    name = fields.String(
        allow_none=True,
        metadata={
            'description': 'The name of the process.',
            'example': 'python',
        },
    )

    parent_pid = fields.Int(
        allow_none=True,
        metadata={
            'description': 'The parent process ID.',
            'example': 1000,
        },
    )

    username = fields.String(
        allow_none=True,
        metadata={
            'description': 'The username of the process owner.',
            'example': 'root',
        },
    )

    command_line = fields.List(
        fields.String(),
        allow_none=True,
        metadata={
            'description': 'The command line arguments of the process.',
            'example': ['python', 'script.py'],
        },
    )

    status = fields.String(
        allow_none=True,
        metadata={
            'description': 'The status of the process.',
            'example': 'running',
        },
    )

    terminal = fields.String(
        allow_none=True,
        metadata={
            'description': 'The terminal of the process.',
            'example': 'tty1',
        },
    )

    current_directory = fields.String(
        allow_none=True,
        metadata={
            'description': 'The current working directory of the process.',
            'example': '/home/user',
        },
    )

    started = DateTime(
        allow_none=True,
        metadata={
            'description': 'The timestamp when the process was started.',
            'example': '2021-01-01T00:00:00+00:00',
        },
    )

    cpu_percent = percent_field(
        missing=0,
        metadata={
            'description': 'The CPU usage percentage of the process, in the range [0, 1].',
            'example': 0.5,
        },
    )

    memory_percent = percent_field(
        missing=0,
        metadata={
            'description': 'The memory usage percentage of the process, in the range [0, 1].',
            'example': 0.5,
        },
    )

    @pre_load
    def pre_load(self, data, **_) -> dict:
        import psutil

        if hasattr(data, 'as_dict'):
            try:
                data = data.as_dict()
            except psutil.NoSuchProcess:
                return {}

        started_ts = data.pop('create_time', None)
        if started_ts is not None:
            data['started'] = datetime.fromtimestamp(started_ts).replace(tzinfo=gettz())

        data['command_line'] = data.pop('cmdline', None)
        data['cpu_percent'] = data.pop('cpu_percent') / 100
        data['current_directory'] = data.pop('cwd', None)
        data['memory_percent'] = data.pop('memory_percent') / 100
        data['parent_pid'] = data.pop('ppid', None)
        return data



import enum

from sqlalchemy import (
    Column,
    DateTime,
    ForeignKey,
    Index,
    Integer,
    String,
    UniqueConstraint,
)

from platypush.common.db import Base


class AuthenticationStatus(enum.Enum):
    """
    Enum for authentication errors.
    """

    OK = ''
    INVALID_AUTH_TYPE = 'invalid_auth_type'
    INVALID_CREDENTIALS = 'invalid_credentials'
    INVALID_METHOD = 'invalid_method'
    INVALID_JWT_TOKEN = 'invalid_jwt_token'
    INVALID_OTP_CODE = 'invalid_otp_code'
    INVALID_SESSION = 'invalid_session'
    MISSING_OTP_CODE = 'missing_otp_code'
    MISSING_PASSWORD = 'missing_password'
    MISSING_USERNAME = 'missing_username'
    PASSWORD_MISMATCH = 'password_mismatch'
    REGISTRATION_DISABLED = 'registration_disabled'
    REGISTRATION_REQUIRED = 'registration_required'
    UNKNOWN_ERROR = 'unknown_error'


class User(Base):
    """Models the User table"""

    __tablename__ = 'user'

    user_id = Column(Integer, primary_key=True, autoincrement=True)
    username = Column(String, unique=True, nullable=False)
    password = Column(String)
    password_salt = Column(String)
    hmac_iterations = Column(Integer)
    created_at = Column(DateTime)


class UserSession(Base):
    """Models the UserSession table"""

    __tablename__ = 'user_session'

    session_id = Column(Integer, primary_key=True, autoincrement=True)
    session_token = Column(String, unique=True, nullable=False)
    csrf_token = Column(String, unique=True)
    user_id = Column(Integer, ForeignKey('user.user_id'), nullable=False)
    created_at = Column(DateTime)
    expires_at = Column(DateTime)


class UserOtp(Base):
    """
    Models the UserOtp table, which contains the OTP secrets for each user.
    """

    __tablename__ = 'user_otp'

    user_id = Column(Integer, ForeignKey('user.user_id'), primary_key=True)
    otp_secret = Column(String, nullable=False, unique=True)
    created_at = Column(DateTime)
    expires_at = Column(DateTime)


class UserBackupCode(Base):
    """
    Models the UserBackupCode table, which contains the backup codes for each
    user with 2FA enabled.
    """

    __tablename__ = 'user_backup_code'

    id = Column(Integer, primary_key=True, autoincrement=True)
    user_id = Column(Integer, ForeignKey('user.user_id'))
    code = Column(String, nullable=False)
    created_at = Column(DateTime)
    expires_at = Column(DateTime)

    UniqueConstraint(user_id, code)


class UserToken(Base):
    """Models the UserToken table"""

    __tablename__ = 'user_token'

    id = Column(Integer, primary_key=True, autoincrement=True)
    user_id = Column(Integer, ForeignKey('user.user_id'), nullable=False)
    token = Column(String, unique=True, nullable=False)
    name = Column(String, nullable=False)
    created_at = Column(DateTime)
    expires_at = Column(DateTime)

    Index('user_token_user_id_token_idx', user_id, token, unique=True)
    Index('user_token_user_id_name_idx', user_id, name, unique=True)
    Index('user_token_user_id_expires_at_idx', user_id, token, expires_at)


# vim:sw=4:ts=4:et:



from typing import List, Optional

from flask import Blueprint, jsonify, request

from platypush.backend.http.app import template_folder
from platypush.backend.http.app.utils import UserAuthStatus, authenticate
from platypush.backend.http.utils import HttpUtils
from platypush.exceptions.user import (
    InvalidCredentialsException,
    InvalidOtpCodeException,
    UserException,
)
from platypush.config import Config
from platypush.context import get_plugin
from platypush.user import UserManager

otp = Blueprint('otp', __name__, template_folder=template_folder)

# Declare routes list
__routes__ = [
    otp,
]


def _get_otp_and_qrcode():
    otp = get_plugin('otp')  # pylint: disable=redefined-outer-name
    qrcode = get_plugin('qrcode')
    assert (
        otp and qrcode
    ), 'The otp and/or qrcode plugins are not available in your installation'

    return otp, qrcode


def _get_username():
    user = HttpUtils.current_user()
    if not user:
        raise InvalidCredentialsException('Invalid user session')

    return str(user.username)


def _get_otp_uri_and_qrcode(username: str, otp_secret: Optional[str] = None):
    if not otp_secret:
        return None, None

    otp, qrcode = _get_otp_and_qrcode()  # pylint: disable=redefined-outer-name
    otp_uri = (
        otp.provision_time_otp(
            name=username,
            secret=otp_secret,
            issuer=f'platypush@{Config.get_device_id()}',
        ).output
        if otp_secret
        else None
    )

    otp_qrcode = (
        qrcode.generate(content=otp_uri, format='png').output.get('data')
        if otp_uri
        else None
    )

    return otp_uri, otp_qrcode


def _verify_code(code: str, otp_secret: str) -> bool:
    otp, _ = _get_otp_and_qrcode()  # pylint: disable=redefined-outer-name
    return otp.verify_time_otp(otp=code, secret=otp_secret).output


def _dump_response(
    username: str,
    otp_secret: Optional[str] = None,
    backup_codes: Optional[List[str]] = None,
):
    otp_uri, otp_qrcode = _get_otp_uri_and_qrcode(username, otp_secret)
    return jsonify(
        {
            'username': username,
            'otp_secret': otp_secret,
            'otp_uri': otp_uri,
            'qrcode': otp_qrcode,
            'backup_codes': backup_codes or [],
        }
    )


def _get_otp():
    username = _get_username()
    user_manager = UserManager()
    otp_secret = user_manager.get_otp_secret(username)
    return _dump_response(
        username=username,
        otp_secret=otp_secret,
    )


def _authenticate_user(username: str, password: Optional[str]):
    assert password, 'The password field is required when setting up OTP'
    user, auth_status = UserManager().authenticate_user(  # type: ignore
        username, password, skip_2fa=True, with_status=True
    )

    if not user:
        raise InvalidCredentialsException(auth_status.value[2])


def _post_otp():
    body = request.json
    assert body, 'Invalid request body'

    username = _get_username()
    dry_run = body.get('dry_run', False)
    otp_secret = body.get('otp_secret')

    if not dry_run:
        _authenticate_user(username, body.get('password'))

        if otp_secret:
            code = body.get('code')
            assert code, 'The code field is required when setting up OTP'

            if not _verify_code(code, otp_secret):
                raise InvalidOtpCodeException()

    user_manager = UserManager()
    user_otp, backup_codes = user_manager.enable_otp(
        username=username,
        otp_secret=otp_secret,
        dry_run=dry_run,
    )

    return _dump_response(
        username=username,
        otp_secret=str(user_otp.otp_secret),
        backup_codes=backup_codes,
    )


def _delete_otp():
    body = request.json
    assert body, 'Invalid request body'

    username = _get_username()
    _authenticate_user(username, body.get('password'))

    user_manager = UserManager()
    user_manager.disable_otp(username)
    return jsonify({'status': 'ok'})


@otp.route('/otp/config', methods=['GET', 'POST', 'DELETE'])
@authenticate()
def otp_route():
    """
    :return: The user's current MFA/OTP configuration:

        .. code-block:: json

            {
                "username": "testuser",
                "otp_secret": "JBSA6ZUZ5DPEK7YV",
                "otp_uri": "otpauth://totp/testuser?secret=JBSA6ZUZ5DPEK7YV&issuer=platypush@localhost",
                "qrcode": "data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAMgAAADICAYAAACtWK6eAAABwklEQVR4nO3dMW7CQBAF0",
                "backup_codes": [
                    "1A2B3C4D5E",
                    "6F7G8H9I0J",
                    "KLMNOPQRST",
                    "UVWXYZ1234",
                    "567890ABCD",
                    "EFGHIJKLMN",
                    "OPQRSTUVWX",
                    "YZ12345678",
                    "90ABCDEF12",
                    "34567890AB"
                ]
            }

    """
    try:
        if request.method.lower() == 'get':
            return _get_otp()

        if request.method.lower() == 'post':
            return _post_otp()

        if request.method.lower() == 'delete':
            return _delete_otp()

        return jsonify({'error': 'Method not allowed'}), 405
    except AssertionError as e:
        return jsonify({'error': str(e)}), 400
    except InvalidCredentialsException:
        return UserAuthStatus.INVALID_CREDENTIALS.to_response()
    except InvalidOtpCodeException:
        return UserAuthStatus.INVALID_OTP_CODE.to_response()
    except UserException as e:
        return jsonify({'error': e.__class__.__name__, 'message': str(e)}), 401
    except Exception as e:
        HttpUtils.log.error(f'Error while processing OTP request: {e}', exc_info=True)
        return jsonify({'error': str(e)}), 500


@otp.route('/otp/refresh-codes', methods=['POST'])
def refresh_codes():
    """
    :return: A new set of backup codes for the user.
    """
    username = _get_username()
    user_manager = UserManager()
    otp_secret = user_manager.get_otp_secret(username)
    if not otp_secret:
        return jsonify({'error': 'OTP not configured for the user'}), 400

    backup_codes = user_manager.refresh_user_backup_codes(username)
    return jsonify({'backup_codes': backup_codes})


# vim:sw=4:ts=4:et:



from typing import Optional
import requests

from platypush.plugins.media._resource import HttpMediaResource, MediaResource
from platypush.utils import wait_for_either

from ._base import DownloadState, DownloadThread, MediaResourceDownloader


class HttpResourceDownloader(MediaResourceDownloader):
    """
    Downloader for generic HTTP URLs.
    """

    def supports(self, resource_type: MediaResource) -> bool:  # type: ignore[override]
        return isinstance(resource_type, HttpMediaResource)

    def download(  # type: ignore[override]
        self,
        resource: HttpMediaResource,
        path: Optional[str] = None,
        timeout: Optional[int] = None,
        **_
    ) -> 'HttpDownloadThread':
        path = path or self.get_download_path(resource=resource)
        download_thread = HttpDownloadThread(
            url=resource.url,
            path=path,
            timeout=timeout,
            on_start=self._media.on_download_start,
            post_event=self._media.post_event,
            stop_event=self._media._should_stop,  # pylint: disable=protected-access
        )

        self._media.start_download(download_thread)
        return download_thread


class HttpDownloadThread(DownloadThread):
    """
    Thread that downloads a generic URL to a file.
    """

    def _run(self):
        interrupted = False

        with requests.get(self.url, timeout=self.timeout, stream=True) as response:
            response.raise_for_status()
            self.size = int(response.headers.get('Content-Length', 0)) or None

            with open(self.path, 'wb') as f:
                self.on_start()

                for chunk in response.iter_content(chunk_size=8192):
                    if not chunk or self.should_stop():
                        interrupted = self.should_stop()
                        if interrupted:
                            self.stop()

                        break

                    self.state = DownloadState.DOWNLOADING
                    f.write(chunk)
                    percent = f.tell() / self.size * 100 if self.size else 0
                    self.progress = percent

                    if self._paused.is_set():
                        wait_for_either(self._downloading, self._stop_event)

        return not interrupted


# vim:sw=4:ts=4:et:



from marshmallow import fields, pre_load

from .._base import SystemBaseSchema


class FanSchema(SystemBaseSchema):
    """
    Base schema for system fan sensors.
    """

    id = fields.String(
        required=True,
        metadata={
            'description': 'Unique ID for the sensor',
            'example': 'acpi_1',
        },
    )

    label = fields.String(
        allow_none=True,
        metadata={
            'description': 'Label for the sensor',
            'example': 'CPU Fan',
        },
    )

    value = fields.Float(
        allow_none=True,
        metadata={
            'description': 'Current fan speed in RPM',
            'example': 1200,
        },
    )

    @pre_load
    def pre_load(self, data: dict, **_) -> dict:
        data = super().pre_load(data)
        data['value'] = data.pop('current', data.pop('value', None))
        return data



from marshmallow import fields, pre_load

from platypush.schemas.dataclasses import percent_field

from .._base import SystemBaseSchema


class MemoryStatsSchema(SystemBaseSchema):
    """
    Base schema for memory stats.
    """

    total = fields.Integer(
        metadata={
            'description': 'Total memory available in bytes.',
            'example': 8589934592,
        }
    )

    available = fields.Integer(
        metadata={
            'description': 'Memory available in bytes.',
            'example': 2147483648,
        }
    )

    used = fields.Integer(
        metadata={
            'description': 'Memory used in bytes.',
            'example': 6442450944,
        }
    )

    free = fields.Integer(
        metadata={
            'description': 'Memory free in bytes.',
            'example': 2147483648,
        }
    )

    active = fields.Integer(
        metadata={
            'description': 'Memory active in bytes.',
            'example': 4294967296,
        }
    )

    inactive = fields.Integer(
        metadata={
            'description': 'Memory inactive in bytes.',
            'example': 2147483648,
        }
    )

    buffers = fields.Integer(
        metadata={
            'description': 'Memory buffers in bytes.',
            'example': 2147483648,
        }
    )

    cached = fields.Integer(
        metadata={
            'description': 'Memory cached in bytes.',
            'example': 2147483648,
        }
    )

    shared = fields.Integer(
        metadata={
            'description': 'Memory shared in bytes.',
            'example': 3221225472,
        }
    )

    percent = percent_field(
        metadata={
            'description': 'Memory usage percentage between 0 and 1.',
            'example': 0.75,
        }
    )

    @pre_load
    def pre_load(self, data: dict, **_) -> dict:
        data = super().pre_load(data)

        # Normalize the percentage between 0 and 1
        if data.get('percent') is not None:
            data['percent'] /= 100
        return data


class SwapStatsSchema(SystemBaseSchema):
    """
    Base schema for swap stats.
    """

    total = fields.Integer(
        metadata={
            'description': 'Total memory available in bytes.',
            'example': 8589934592,
        }
    )

    used = fields.Integer(
        metadata={
            'description': 'Memory used in bytes.',
            'example': 6442450944,
        }
    )

    free = fields.Integer(
        metadata={
            'description': 'Memory free in bytes.',
            'example': 2147483648,
        }
    )

    percent = percent_field(
        metadata={
            'description': 'Memory usage percentage between 0 and 1.',
            'example': 0.75,
        }
    )

    @pre_load
    def pre_load(self, data: dict, **_) -> dict:
        data = super().pre_load(data)

        # Normalize the percentage between 0 and 1
        if data.get('percent') is not None:
            data['percent'] /= 100
        return data



from sqlalchemy import Boolean, Column, Float, ForeignKey, Integer, JSON, String

from . import Entity
from .devices import Device
from .sensors import NumericSensor, PercentSensor
from .temperature import TemperatureSensor


class Cpu(Entity):
    """
    ``CPU`` ORM (container) model.
    """

    __tablename__ = 'cpu'

    id = Column(Integer, ForeignKey(Entity.id, ondelete='CASCADE'), primary_key=True)

    percent = Column(Float)

    __table_args__ = {'extend_existing': True}
    __mapper_args__ = {
        'polymorphic_identity': __tablename__,
    }


class CpuInfo(Entity):
    """
    ``CpuInfo`` ORM model.
    """

    __tablename__ = 'cpu_info'

    id = Column(Integer, ForeignKey(Entity.id, ondelete='CASCADE'), primary_key=True)

    architecture = Column(String)
    bits = Column(Integer)
    cores = Column(Integer)
    vendor = Column(String)
    brand = Column(String)
    frequency_advertised = Column(Integer)
    frequency_actual = Column(Integer)
    flags = Column(JSON)
    l1_instruction_cache_size = Column(Integer)
    l1_data_cache_size = Column(Integer)
    l2_cache_size = Column(Integer)
    l3_cache_size = Column(Integer)

    __table_args__ = {'extend_existing': True}
    __mapper_args__ = {
        'polymorphic_identity': __tablename__,
    }


class CpuTimes(Entity):
    """
    ``CpuTimes`` ORM (container) model.
    """

    __tablename__ = 'cpu_times'

    id = Column(Integer, ForeignKey(Entity.id, ondelete='CASCADE'), primary_key=True)

    __table_args__ = {'extend_existing': True}
    __mapper_args__ = {
        'polymorphic_identity': __tablename__,
    }


class CpuStats(Entity):
    """
    ``CpuStats`` ORM (container) model.
    """

    __tablename__ = 'cpu_stats'

    id = Column(Integer, ForeignKey(Entity.id, ondelete='CASCADE'), primary_key=True)

    __table_args__ = {'extend_existing': True}
    __mapper_args__ = {
        'polymorphic_identity': __tablename__,
    }


class MemoryStats(Entity):
    """
    ``MemoryStats`` ORM model.
    """

    __tablename__ = 'memory_stats'

    id = Column(Integer, ForeignKey(Entity.id, ondelete='CASCADE'), primary_key=True)

    total = Column(Integer)
    available = Column(Integer)
    used = Column(Integer)
    free = Column(Integer)
    active = Column(Integer)
    inactive = Column(Integer)
    buffers = Column(Integer)
    cached = Column(Integer)
    shared = Column(Integer)
    percent = Column(Float)

    __table_args__ = {'extend_existing': True}
    __mapper_args__ = {
        'polymorphic_identity': __tablename__,
    }


class SwapStats(Entity):
    """
    ``SwapStats`` ORM model.
    """

    __tablename__ = 'swap_stats'

    id = Column(Integer, ForeignKey(Entity.id, ondelete='CASCADE'), primary_key=True)

    total = Column(Integer)
    used = Column(Integer)
    free = Column(Integer)
    percent = Column(Float)

    __table_args__ = {'extend_existing': True}
    __mapper_args__ = {
        'polymorphic_identity': __tablename__,
    }


class Disk(Entity):
    """
    ``Disk`` ORM model.
    """

    __tablename__ = 'disk'

    id = Column(Integer, ForeignKey(Entity.id, ondelete='CASCADE'), primary_key=True)

    mountpoint = Column(String)
    fstype = Column(String)
    opts = Column(String)
    total = Column(Integer)
    used = Column(Integer)
    free = Column(Integer)
    percent = Column(Float)
    read_count = Column(Integer)
    write_count = Column(Integer)
    read_bytes = Column(Integer)
    write_bytes = Column(Integer)
    read_time = Column(Float)
    write_time = Column(Float)
    busy_time = Column(Float)

    __table_args__ = {'extend_existing': True}
    __mapper_args__ = {
        'polymorphic_identity': __tablename__,
    }


class NetworkInterface(Device):
    """
    ``NetworkInterface`` ORM model.
    """

    __tablename__ = 'network_interface'

    id = Column(Integer, ForeignKey(Device.id, ondelete='CASCADE'), primary_key=True)

    bytes_sent = Column(Integer)
    bytes_recv = Column(Integer)
    packets_sent = Column(Integer)
    packets_recv = Column(Integer)
    errors_in = Column(Integer)
    errors_out = Column(Integer)
    drop_in = Column(Integer)
    drop_out = Column(Integer)
    addresses = Column(JSON)
    speed = Column(Integer)
    mtu = Column(Integer)
    duplex = Column(String)
    flags = Column(JSON)

    __table_args__ = {'extend_existing': True}
    __mapper_args__ = {
        'polymorphic_identity': __tablename__,
    }


class SystemTemperature(TemperatureSensor):
    """
    Extends the ``TemperatureSensor``.
    """

    __tablename__ = 'system_temperature'

    id = Column(
        Integer,
        ForeignKey(TemperatureSensor.id, ondelete='CASCADE'),
        primary_key=True,
    )

    high = Column(Float)
    critical = Column(Float)

    __table_args__ = {'extend_existing': True}
    __mapper_args__ = {
        'polymorphic_identity': __tablename__,
    }


class SystemFan(NumericSensor):
    """
    ``SystemFan`` ORM model.
    """

    __tablename__ = 'system_fan'

    id = Column(
        Integer,
        ForeignKey(NumericSensor.id, ondelete='CASCADE'),
        primary_key=True,
    )

    __table_args__ = {'extend_existing': True}
    __mapper_args__ = {
        'polymorphic_identity': __tablename__,
    }


class SystemBattery(PercentSensor):
    """
    ``SystemBattery`` ORM model.
    """

    __tablename__ = 'system_battery'

    id = Column(
        Integer,
        ForeignKey(PercentSensor.id, ondelete='CASCADE'),
        primary_key=True,
    )

    seconds_left = Column(Float)
    power_plugged = Column(Boolean)

    __table_args__ = {'extend_existing': True}
    __mapper_args__ = {
        'polymorphic_identity': __tablename__,
    }



from dataclasses import dataclass

from ._base import MediaResource


@dataclass
class HttpMediaResource(MediaResource):
    """
    Models a media resource that is read from an HTTP response.
    """

    def open(self, *args, **kwargs):
        return super().open(*args, **kwargs)

    def close(self) -> None:
        super().close()


# vim:sw=4:ts=4:et:



from ._base import MediaResourceParser
from .file import FileResourceParser
from .http import HttpResourceParser
from .torrent import TorrentResourceParser
from .youtube import YoutubeResourceParser


parsers = [
    FileResourceParser,
    YoutubeResourceParser,
    TorrentResourceParser,
    HttpResourceParser,
]


__all__ = [
    'MediaResourceParser',
    'FileResourceParser',
    'HttpResourceParser',
    'TorrentResourceParser',
    'YoutubeResourceParser',
    'parsers',
]


# vim:sw=4:ts=4:et:



import os
import pathlib
from contextlib import contextmanager
from datetime import datetime as dt
from typing import IO, Optional, Tuple

from tornado.web import stream_request_body

from platypush.utils import get_mime_type

from .. import StreamingRoute


@stream_request_body
class FileRoute(StreamingRoute):
    """
    Generic route to read the content of a file on the server.
    """

    BUFSIZE = 1024
    _bytes_written = 0
    _out_f: Optional[IO[bytes]] = None

    @classmethod
    def path(cls) -> str:
        """
        Route: GET /file?path=<path>[&download]
        """
        return r"^/file$"

    @property
    def download(self) -> bool:
        return 'download' in self.request.arguments

    @property
    def file_path(self) -> str:
        return os.path.expanduser(
            self.request.arguments.get('path', [b''])[0].decode('utf-8')
        )

    @property
    def file_size(self) -> int:
        return os.path.getsize(self.file_path)

    @property
    def _content_length(self) -> int:
        return int(self.request.headers.get('Content-Length', 0))

    @property
    def range(self) -> Tuple[Optional[int], Optional[int]]:
        range_hdr = self.request.headers.get('Range')
        if not range_hdr:
            return None, None

        start, end = range_hdr.split('=')[-1].split('-')
        start = int(start) if start else 0
        end = int(end) if end else self.file_size - 1
        return start, end

    def set_headers(self):
        self.set_header('Content-Length', str(os.path.getsize(self.file_path)))
        self.set_header(
            'Content-Type', get_mime_type(self.file_path) or 'application/octet-stream'
        )
        self.set_header('Accept-Ranges', 'bytes')
        self.set_header(
            'Last-Modified',
            dt.fromtimestamp(os.path.getmtime(self.file_path)).strftime(
                '%a, %d %b %Y %H:%M:%S GMT'
            ),
        )

        if self.download:
            self.set_header(
                'Content-Disposition',
                f'attachment; filename="{os.path.basename(self.file_path)}"',
            )

        if self.range[0] is not None:
            start, end = self.range
            self.set_header(
                'Content-Range',
                f'bytes {start}-{end}/{self.file_size}',
            )
            self.set_status(206)

    @contextmanager
    def _serve(self):
        path = self.file_path
        if not path:
            self.write_error(400, 'Missing path argument')
            return

        self.logger.debug('Received file read request for %r', path)

        try:
            with open(path, 'rb') as f:
                self.set_headers()
                yield f
        except FileNotFoundError:
            self.write_error(404, 'File not found')
            yield
            return
        except PermissionError:
            self.write_error(403, 'Permission denied')
            yield
            return
        except Exception as e:
            self.write_error(500, str(e))
            yield
            return

        self.finish()

    def on_finish(self) -> None:
        if self._out_f:
            try:
                if not (self._out_f and self._out_f.closed):
                    self._out_f.close()
            except Exception as e:
                self.logger.warning('Error while closing the output file: %s', e)

            self._out_f = None

        return super().on_finish()

    def _validate_upload(self, force: bool = False) -> bool:
        if not self.file_path:
            self.write_error(400, 'Missing path argument')
            return False

        if not self._out_f:
            if not force and os.path.exists(self.file_path):
                self.write_error(409, f'{self.file_path} already exists')
                return False

            self._bytes_written = 0
            dir_path = os.path.dirname(self.file_path)

            try:
                pathlib.Path(dir_path).mkdir(parents=True, exist_ok=True)
                self._out_f = open(  # pylint: disable=consider-using-with
                    self.file_path, 'wb'
                )
            except PermissionError:
                self.write_error(403, 'Permission denied')
                return False

        return True

    def finish(self, *args, **kwargs):  # type: ignore
        try:
            return super().finish(*args, **kwargs)
        except Exception as e:
            self.logger.warning('Error while finishing the request: %s', e)

    def data_received(self, chunk: bytes):
        # Ignore unless we're in POST/PUT mode
        if self.request.method not in ('POST', 'PUT'):
            return

        force = self.request.method == 'PUT'
        if not self._validate_upload(force=force):
            self.finish()
            return

        if not chunk:
            self.logger.debug('Received EOF from client')
            self.finish()
            return

        assert self._out_f
        self._out_f.write(chunk)
        self._out_f.flush()
        self._bytes_written += len(chunk)
        self.logger.debug(
            'Written chunk of size %d to %s, progress: %d/%d',
            len(chunk),
            self.file_path,
            self._bytes_written,
            self._content_length,
        )

        self.flush()

    def get(self) -> None:
        with self._serve() as f:
            if f:
                while True:
                    chunk = f.read(self.BUFSIZE)
                    if not chunk:
                        break

                    self.write(chunk)
                    self.flush()

    def head(self) -> None:
        with self._serve():
            pass

    def post(self) -> None:
        self.logger.info('Receiving file POST upload request for %r', self.file_path)

    def put(self) -> None:
        self.logger.info('Receiving file PUT upload request for %r', self.file_path)



import io
import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass, field, fields
from subprocess import Popen
from typing import IO, Iterable, Optional

from platypush.plugins import Plugin


@dataclass
class MediaResource(ABC):
    """
    Models a generic media resource.

    In this case resource/URL can be passed directly to the player.
    """

    resource: str
    url: str
    media_plugin: Plugin
    fd: Optional[IO] = None
    title: Optional[str] = None
    ext: Optional[str] = None
    description: Optional[str] = None
    filename: Optional[str] = None
    size: Optional[int] = None
    image: Optional[str] = None
    duration: Optional[float] = None
    channel: Optional[str] = None
    channel_id: Optional[str] = None
    channel_url: Optional[str] = None
    type: Optional[str] = None
    width: Optional[int] = None
    height: Optional[int] = None
    resolution: Optional[str] = None
    timestamp: Optional[float] = None
    fps: Optional[float] = None
    audio_channels: Optional[int] = None
    view_count: Optional[int] = None
    categories: Optional[Iterable[str]] = field(default_factory=list)
    tags: Optional[Iterable[str]] = field(default_factory=list)

    @property
    def _logger(self):
        return logging.getLogger(self.__class__.__name__)

    @property
    def _media(self):
        """
        This workaround is required to avoid circular imports.
        """
        from platypush.plugins.media import MediaPlugin

        assert isinstance(self.media_plugin, MediaPlugin)
        return self.media_plugin

    @abstractmethod
    def open(self, *_, **__) -> IO:
        """
        Opens the media resource.
        """
        if self.fd is not None:
            try:
                self.fd.seek(0)
            except io.UnsupportedOperation:
                pass

        return self.fd

    def close(self) -> None:
        """
        Closes the media resource.
        """
        if self.fd is not None:
            self.fd.close()
            self.fd = None

    def __enter__(self) -> IO:
        """
        Opens the media resource using a context manager.
        """
        return self.open()

    def __exit__(self, *_, **__) -> None:
        """
        Closes the media resource using a context manager.
        """
        self.close()

    def to_dict(self) -> dict:
        """
        Converts the media resource to a dictionary ready to be serialized.
        """
        return {
            f.name: getattr(self, f.name)
            for f in fields(self)
            if f.name not in ['media_plugin', 'fd']
        }


@dataclass
class PopenMediaResource(MediaResource, ABC):
    """
    Models a media resource that is read from a Popen object.
    """

    proc: Optional[Popen] = None

    def close(self) -> None:
        """
        Closes the media resource.
        """
        if self.proc is not None:
            self.proc.terminate()
            self.proc.wait(1)
            if self.proc and self.proc.poll() is None:
                self.proc.kill()
                self.proc.wait(1)

            self.proc = None

        super().close()

    def to_dict(self) -> dict:
        """
        Converts the media resource to a dictionary ready to be serialized.
        """
        ret = super().to_dict()
        ret.pop('proc', None)
        return ret


# vim:sw=4:ts=4:et:



from dataclasses import dataclass
from typing import Optional


@dataclass
class Temperature:
    """
    System temperature sensor data class.
    """

    id: str
    label: Optional[str] = None
    value: Optional[float] = None
    high: Optional[float] = None
    critical: Optional[float] = None



from dataclasses import dataclass, field
from typing import List, Optional, Tuple


@dataclass
class CpuInfo:
    """
    CPU info data class.
    """

    bits: int
    cores: int
    architecture: Optional[str] = None
    vendor: Optional[str] = None
    brand: Optional[str] = None
    frequency_advertised: Optional[float] = None
    frequency_actual: Optional[float] = None
    flags: List[str] = field(default_factory=list)
    l1_instruction_cache_size: Optional[float] = None
    l1_data_cache_size: Optional[float] = None
    l2_cache_size: Optional[float] = None
    l3_cache_size: Optional[float] = None


@dataclass
class CpuTimes:
    """
    CPU times data class.
    """

    user: Optional[float] = None
    nice: Optional[float] = None
    system: Optional[float] = None
    idle: Optional[float] = None
    iowait: Optional[float] = None
    irq: Optional[float] = None
    softirq: Optional[float] = None
    steal: Optional[float] = None
    guest: Optional[float] = None
    guest_nice: Optional[float] = None


@dataclass
class CpuStats:
    """
    CPU stats data class.
    """

    ctx_switches: int
    interrupts: int
    soft_interrupts: int
    syscalls: int


@dataclass
class CpuFrequency:
    """
    CPU frequency data class.
    """

    current: float
    min: float
    max: float


@dataclass
class Cpu:
    """
    CPU data aggregate dataclass.
    """

    info: CpuInfo
    times: CpuTimes
    frequency: CpuFrequency
    stats: CpuStats
    load_avg: Tuple[float, float, float]
    percent: float



import logging
import threading
import time

from platypush.bus import Bus
from platypush.message import Message

logger = logging.getLogger('platypush:bus:redis')


class RedisBus(Bus):
    """
    Overrides the in-process in-memory local bus with a Redis bus
    """

    DEFAULT_REDIS_QUEUE: str = 'platypush/bus'

    def __init__(self, *_, on_message=None, redis_queue=None, **kwargs):
        super().__init__(on_message=on_message)
        self.redis_args = kwargs
        self._redis = None
        self.redis_queue = redis_queue or self.DEFAULT_REDIS_QUEUE
        self.on_message = on_message
        self.thread_id = threading.get_ident()
        self._pubsub = None
        self._pubsub_lock = threading.RLock()

    @property
    def redis(self):
        from platypush.utils import get_redis

        if not self._redis:
            self._redis = get_redis(**self.redis_args)
        return self._redis

    @property
    def pubsub(self):
        with self._pubsub_lock:
            if not self._pubsub:
                self._pubsub = self.redis.pubsub()
            return self._pubsub

    def poll(self):
        """
        Polls the Redis queue for new messages
        """
        from redis.exceptions import ConnectionError as RedisConnectionError

        from platypush.message.event.application import ApplicationStartedEvent
        from platypush.utils import redis_pools

        has_error = False

        while not self.should_stop():
            with self.pubsub as pubsub:
                try:
                    pubsub.subscribe(self.redis_queue)
                    self.post(ApplicationStartedEvent())

                    for msg in pubsub.listen():
                        if has_error:
                            logger.info('Redis connection restored')

                        has_error = False
                        if msg.get('type') != 'message':
                            continue

                        if self.should_stop():
                            break

                        try:
                            data = msg.get('data', b'').decode('utf-8')
                            logger.debug('Received message on the Redis bus: %r', data)
                            parsed_msg = Message.build(data)
                            if parsed_msg and self.on_message:
                                self.on_message(parsed_msg)
                        except Exception as e:
                            logger.exception(e)
                except RedisConnectionError as e:
                    if not (self.should_stop() or has_error):
                        logger.warning('Redis connection error: %s', e)

                    has_error = True
                    pubsub.close()
                    redis_pools.clear()  # Clear the connection pool
                    self._redis = None
                    time.sleep(1)
                finally:
                    try:
                        pubsub.unsubscribe(self.redis_queue)
                    except RedisConnectionError:
                        pass

    def post(self, msg):
        """
        Sends a message to the Redis queue
        """
        from redis import exceptions

        try:
            self.redis.publish(self.redis_queue, str(msg))
        except exceptions.ConnectionError as e:
            if not self.should_stop():
                # Raise the exception only if the bus it not supposed to be
                # stopped
                raise e

    def stop(self):
        super().stop()
        self.redis.close()


# vim:sw=4:ts=4:et:



from enum import Enum

from marshmallow import fields, pre_load

from platypush.schemas.dataclasses import percent_field

from .._base import SystemBaseSchema


class BatterySchema(SystemBaseSchema):
    """
    System battery sensor schema.
    """

    seconds_left = fields.Float(
        allow_none=True,
        metadata={
            'description': 'Number of seconds left before the battery runs out',
            'example': 7200,
        },
    )

    power_plugged = fields.Boolean(
        metadata={
            'description': 'Whether the battery is currently plugged in',
            'example': False,
        }
    )

    value = percent_field(
        allow_none=True,
        metadata={
            'description': 'Current battery charge level, as a percentage between 0 and 1',
            'example': 0.5,
        },
    )

    @pre_load
    def pre_load(self, data: dict, **_) -> dict:
        data = super().pre_load(data)
        percent = data.pop('percent', data.pop('value', None))
        seconds_left = data.pop('secsleft', data.pop('seconds_left', None))
        data['value'] = percent / 100 if percent is not None else None
        data['seconds_left'] = None if isinstance(seconds_left, Enum) else seconds_left
        return data



from dataclasses import dataclass, field
from socket import AddressFamily
from typing import List, Optional


@dataclass
class NetworkInterfaceAddress:
    """
    Network interface address data class.
    """

    family: Optional[AddressFamily] = None
    address: Optional[str] = None
    netmask: Optional[str] = None
    broadcast: Optional[str] = None


@dataclass
class NetworkInterface:
    """
    Network interface statistics data class.
    """

    interface: Optional[str] = None
    bytes_sent: int = 0
    bytes_recv: int = 0
    packets_sent: int = 0
    packets_recv: int = 0
    errors_in: int = 0
    errors_out: int = 0
    drop_in: int = 0
    drop_out: int = 0
    is_up: bool = False
    speed: int = 0
    mtu: int = 0
    duplex: Optional[str] = None
    flags: List[str] = field(default_factory=list)
    addresses: List[NetworkInterfaceAddress] = field(default_factory=list)



import base64
import datetime
import hashlib
import json
import os
import random
import time
from typing import List, Optional, Tuple, Union

import rsa
from sqlalchemy import and_, or_
from sqlalchemy.orm import make_transient

from platypush.common.db import Base
from platypush.config import Config
from platypush.context import get_plugin
from platypush.exceptions.user import (
    InvalidCredentialsException,
    InvalidJWTTokenException,
    InvalidTokenException,
    NoUserException,
    OtpRecordAlreadyExistsException,
)
from platypush.utils import get_or_generate_stored_rsa_key_pair, utcnow

from ._model import (
    AuthenticationStatus,
    User,
    UserBackupCode,
    UserOtp,
    UserSession,
    UserToken,
)


class UserManager:
    """
    Main class for managing platform users
    """

    _otp_workdir = os.path.join(Config.get_workdir(), 'otp')
    _otp_keyfile = os.path.join(_otp_workdir, 'key')
    _otp_keyfile_pub = f'{_otp_keyfile}.pub'

    _jwt_workdir = os.path.join(Config.get_workdir(), 'jwt')
    _jwt_keyfile = os.path.join(_jwt_workdir, 'id_rsa')
    _jwt_keyfile_pub = f'{_jwt_keyfile}.pub'

    def __init__(self):
        db_plugin = get_plugin('db')
        assert db_plugin, 'Database plugin not configured'
        self.db = db_plugin
        self.db.create_all(self.db.get_engine(), Base)

    @staticmethod
    def _mask_password(user):
        make_transient(user)
        user.password = None
        return user

    def _get_session(self, *args, **kwargs):
        return self.db.get_session(self.db.get_engine(), *args, **kwargs)

    @classmethod
    def _get_jwt_rsa_key_pair(cls):
        """
        Get or generate the JWT RSA key pair.
        """
        return get_or_generate_stored_rsa_key_pair(cls._jwt_keyfile, size=2048)

    @classmethod
    def _get_or_generate_otp_rsa_key_pair(cls):
        """
        Get or generate the OTP RSA key pair.
        """
        return get_or_generate_stored_rsa_key_pair(cls._otp_keyfile, size=2048)

    @staticmethod
    def _encrypt(data: Union[str, bytes, dict, list, tuple], key: rsa.PublicKey) -> str:
        """
        Encrypt the data using the given RSA public key.
        """
        if isinstance(data, tuple):
            data = list(data)
        if isinstance(data, (dict, list)):
            data = json.dumps(data, sort_keys=True, indent=None)
        if isinstance(data, str):
            data = data.encode('ascii')

        return base64.b64encode(rsa.encrypt(data, key)).decode()

    @staticmethod
    def _decrypt(data: Union[str, bytes], key: rsa.PrivateKey) -> str:
        """
        Decrypt the data using the given RSA private key.
        """
        if isinstance(data, str):
            data = data.encode('ascii')

        return rsa.decrypt(base64.b64decode(data), key).decode()

    def get_user(self, username):
        with self._get_session() as session:
            user = self._get_user(session, username)
            if not user:
                return None

            session.expunge(user)
            return self._mask_password(user)

    def get_user_count(self):
        with self._get_session() as session:
            return session.query(User).count()

    def get_users(self):
        with self._get_session() as session:
            return session.query(User).all()

    def create_user(self, username: str, password: str, **kwargs):
        if not username:
            raise ValueError('Invalid or empty username')
        if not password:
            raise ValueError('Please provide a password for the user')

        with self._get_session(locked=True) as session:
            user = self._get_user(session, username)
            if user:
                raise NameError(f'The user {username} already exists')

            password_salt = os.urandom(16)
            hmac_iterations = 100_000
            record = User(
                username=username,
                password=self._encrypt_password(
                    password, password_salt, hmac_iterations
                ),
                password_salt=password_salt.hex(),
                hmac_iterations=hmac_iterations,
                created_at=utcnow(),
                **kwargs,
            )

            session.add(record)
            session.commit()
            user = self._get_user(session, username)

        return self._mask_password(user)

    def update_password(self, username, old_password, new_password, code=None):
        with self._get_session(locked=True) as session:
            if not self._authenticate_user(session, username, old_password, code=code):
                return False

            user = self._get_user(session, username)
            user.password_salt = user.password_salt or os.urandom(16).hex()
            user.hmac_iterations = user.hmac_iterations or 100_000
            salt = bytes.fromhex(user.password_salt)
            user.password = self._encrypt_password(
                new_password, salt, user.hmac_iterations
            )
            session.commit()
            return True

    def authenticate_user(
        self, username, password, code=None, skip_2fa=False, with_status=False
    ):
        with self._get_session() as session:
            return self._authenticate_user(
                session,
                username,
                password,
                code=code,
                skip_2fa=skip_2fa,
                with_status=with_status,
            )

    def authenticate_user_session(self, session_token, with_status=False):
        with self._get_session() as session:
            users_count = session.query(User).count()
            if not users_count:
                return (
                    (None, None, AuthenticationStatus.REGISTRATION_REQUIRED)
                    if with_status
                    else (None, None)
                )

            user_session = (
                session.query(UserSession)
                .filter_by(session_token=session_token)
                .first()
            )

            expires_at = (
                user_session.expires_at.replace(tzinfo=datetime.timezone.utc)
                if user_session and user_session.expires_at
                else None
            )

            if not user_session or (expires_at and expires_at < utcnow()):
                return (
                    (None, None, AuthenticationStatus.INVALID_CREDENTIALS)
                    if with_status
                    else (None, None)
                )

            user = session.query(User).filter_by(user_id=user_session.user_id).first()
            return (
                (self._mask_password(user), user_session, AuthenticationStatus.OK)
                if with_status
                else (
                    self._mask_password(user),
                    user_session,
                )
            )

    def delete_user(self, username):
        with self._get_session(locked=True) as session:
            user = self._get_user(session, username)
            if not user:
                raise NameError(f'No such user: {username}')

            user_sessions = (
                session.query(UserSession).filter_by(user_id=user.user_id).all()
            )
            for user_session in user_sessions:
                session.delete(user_session)

            session.delete(user)
            session.commit()
            return True

    def delete_user_session(self, session_token):
        with self._get_session(locked=True) as session:
            user_session = (
                session.query(UserSession)
                .filter_by(session_token=session_token)
                .first()
            )

            if not user_session:
                return False

            session.delete(user_session)
            session.commit()
            return True

    def create_user_session(
        self,
        username,
        password,
        code=None,
        expires_at=None,
        with_status=False,
    ):
        with self._get_session(locked=True) as session:
            user, status = self._authenticate_user(  # type: ignore
                session,
                username,
                password,
                code=code,
                with_status=with_status,
            )

            if not user:
                return None if not with_status else (None, status)

            if expires_at:
                if isinstance(expires_at, (int, float)):
                    expires_at = datetime.datetime.fromtimestamp(expires_at)
                elif isinstance(expires_at, str):
                    expires_at = datetime.datetime.fromisoformat(expires_at)

            user_session = UserSession(
                user_id=user.user_id,
                session_token=self.generate_session_token(),
                csrf_token=self.generate_session_token(),
                created_at=utcnow(),
                expires_at=expires_at,
            )

            session.add(user_session)
            session.commit()
            return user_session, (
                AuthenticationStatus.OK if not with_status else status
            )

    def create_otp_secret(
        self,
        username: str,
        expires_at: Optional[datetime.datetime] = None,
        otp_secret: Optional[str] = None,
        dry_run: bool = False,
    ):
        # Generate a new OTP secret and encrypt it with the OTP RSA key pair
        otp_secret = otp_secret or "".join(
            random.choice("ABCDEFGHIJKLMNOPQRSTUVWXYZ234567") for _ in range(32)
        )

        with self._get_session(locked=True) as session:
            user = self._get_user(session, username)
            if not user:
                raise InvalidCredentialsException()

            # Create a new OTP secret
            user_otp = UserOtp(
                user_id=user.user_id,
                otp_secret=otp_secret,
                created_at=utcnow(),
                expires_at=expires_at,
            )

            if not dry_run:
                # Store a copy of the OTP secret encrypted with the RSA public key
                pubkey, _ = self._get_or_generate_otp_rsa_key_pair()
                encrypted_secret = self._encrypt(otp_secret, pubkey)
                encrypted_otp = UserOtp(
                    user_id=user_otp.user_id,
                    otp_secret=encrypted_secret,
                    created_at=user_otp.created_at,
                    expires_at=user_otp.expires_at,
                )

                # Remove any existing OTP secret and replace it with the new one
                session.query(UserOtp).filter_by(user_id=user.user_id).delete()
                session.add(encrypted_otp)
                session.commit()

        return user_otp

    def get_otp_secret(self, username: str) -> Optional[str]:
        with self._get_session() as session:
            user = self._get_user(session, username)
            if not user:
                return None

            user_otp = session.query(UserOtp).filter_by(user_id=user.user_id).first()
            if not user_otp:
                return None

            _, priv_key = self._get_or_generate_otp_rsa_key_pair()
            return self._decrypt(user_otp.otp_secret, priv_key)

    @staticmethod
    def _get_user(session, username):
        return session.query(User).filter_by(username=username).first()

    @classmethod
    def _encrypt_password(
        cls,
        pwd: str,
        salt: Optional[Union[str, bytes]] = None,
        iterations: Optional[int] = None,
    ) -> str:
        # Legacy password check that uses bcrypt if no salt and iterations are provided
        # See https://git.platypush.tech/platypush/platypush/issues/397
        if not (salt and iterations):
            import bcrypt

            return bcrypt.hashpw(pwd.encode(), bcrypt.gensalt(12)).decode()

        if isinstance(salt, str):
            salt = bytes.fromhex(salt)

        return hashlib.pbkdf2_hmac('sha256', pwd.encode(), salt, iterations).hex()

    @classmethod
    def _check_password(
        cls,
        pwd: str,
        hashed_pwd: str,
        salt: Optional[bytes] = None,
        iterations: Optional[int] = None,
    ) -> bool:
        # Legacy password check that uses bcrypt if no salt and iterations are provided
        # See https://git.platypush.tech/platypush/platypush/issues/397
        if not (salt and iterations):
            import bcrypt

            return bcrypt.checkpw(pwd.encode(), hashed_pwd.encode())

        return (
            hashlib.pbkdf2_hmac(
                'sha256',
                pwd.encode(),
                salt,
                iterations,
            ).hex()
            == hashed_pwd
        )

    @staticmethod
    def _to_bytes(data) -> bytes:
        if isinstance(data, str):
            data = data.encode()
        return data

    @staticmethod
    def generate_session_token():
        rand = bytes(random.randint(0, 255) for _ in range(0, 255))
        return hashlib.sha256(rand).hexdigest()

    def get_user_by_session(self, session_token: str):
        """
        Get a user associated to a session token.

        :param session_token: Session token.
        """
        with self._get_session() as session:
            return (
                session.query(User)
                .join(UserSession)
                .filter_by(session_token=session_token)
                .first()
            )

    def generate_jwt_token(
        self,
        username: str,
        password: str,
        expires_at: Optional[datetime.datetime] = None,
    ) -> str:
        """
        Create a user JWT token for API usage.

        :param username: User name.
        :param password: Password.
        :param expires_at: Expiration datetime of the token.
        :return: The generated JWT token as a string.
        :raises: :class:`platypush.exceptions.user.InvalidCredentialsException` in case of invalid credentials.
        """
        user = self.authenticate_user(username, password, skip_2fa=True)
        if not user:
            raise InvalidCredentialsException()

        pub_key, _ = self._get_jwt_rsa_key_pair()
        return self._encrypt(
            {
                'username': username,
                'password': password,
                'created_at': datetime.datetime.now().timestamp(),
                'expires_at': expires_at.timestamp() if expires_at else None,
            },
            pub_key,
        )

    def validate_jwt_token(self, token: str) -> User:
        """
        Validate a JWT token.

        :param token: Token to validate.
        :return: On success, it returns the user associated to the token.
        :raises: :class:`platypush.exceptions.user.InvalidJWTTokenException` in
            case of invalid token.
        :raises: :class:`platypush.exceptions.user.InvalidCredentialsException`
            in case of invalid credentials stored in the token.
        """
        _, priv_key = self._get_jwt_rsa_key_pair()

        try:
            payload = json.loads(self._decrypt(token, priv_key))
        except (TypeError, ValueError) as e:
            raise InvalidJWTTokenException(f'Could not decode JWT token: {e}') from e

        expires_at = payload.get('expires_at')
        if expires_at and time.time() > expires_at:
            raise InvalidJWTTokenException('Expired JWT token')

        user = self.authenticate_user(
            payload.get('username', ''), payload.get('password', ''), skip_2fa=True
        )

        if not user:
            raise InvalidCredentialsException()

        return user

    def _authenticate_user(
        self,
        session,
        username: str,
        password: str,
        code: Optional[str] = None,
        skip_2fa: bool = False,
        with_status: bool = False,
    ) -> Union[Optional['User'], Tuple[Optional['User'], 'AuthenticationStatus']]:
        """
        :return: :class:`platypush.user.User` instance if the user exists and
        the password is valid, ``None`` otherwise.
        """
        user = self._get_user(session, username)

        # The user does not exist
        if not user:
            return (
                None
                if not with_status
                else (None, AuthenticationStatus.INVALID_CREDENTIALS)
            )

        # The password is not correct
        if not self._check_password(
            password,
            user.password,
            bytes.fromhex(user.password_salt) if user.password_salt else None,
            user.hmac_iterations,
        ):
            return (
                None
                if not with_status
                else (None, AuthenticationStatus.INVALID_CREDENTIALS)
            )

        otp_secret = self.get_otp_secret(username)

        # The user doesn't have 2FA enabled and the password is correct:
        # authentication successful
        if skip_2fa or not otp_secret:
            return user if not with_status else (user, AuthenticationStatus.OK)

        # The user has 2FA enabled but the code is missing
        if not code:
            return (
                None
                if not with_status
                else (None, AuthenticationStatus.MISSING_OTP_CODE)
            )

        # The user has 2FA enabled and a TOTP code is provided
        if self.validate_otp_code(username, code):
            return user if not with_status else (user, AuthenticationStatus.OK)

        # The user has 2FA enabled and a backup code is provided
        if not self.validate_backup_code(username, code):
            return (
                None
                if not with_status
                else (None, AuthenticationStatus.INVALID_OTP_CODE)
            )

        return user if not with_status else (user, AuthenticationStatus.OK)

    def refresh_user_backup_codes(self, username: str) -> List[str]:
        """
        Refresh the backup codes for a user with 2FA enabled.
        """
        backup_codes = [
            "".join(
                random.choice("ABCDEFGHIJKLMNOPQRSTUVWXYZ234567") for _ in range(16)
            )
            for _ in range(10)
        ]

        with self._get_session(locked=True) as session:
            user = self._get_user(session, username)
            if not user:
                return []

            session.query(UserBackupCode).filter_by(user_id=user.user_id).delete()
            stored_codes = []

            for backup_code in backup_codes:
                encrypted_code = self._encrypt_password(
                    backup_code,
                    salt=user.password_salt,
                    iterations=user.hmac_iterations,
                )

                user_backup_code = UserBackupCode(
                    user_id=user.user_id,
                    code=encrypted_code,
                    created_at=utcnow(),
                    expires_at=utcnow() + datetime.timedelta(days=30),
                )

                session.add(user_backup_code)
                stored_codes.append(backup_code)

            session.commit()
            return stored_codes

    def validate_backup_code(self, username: str, code: str) -> bool:
        with self._get_session() as session:
            user = self._get_user(session, username)
            if not user:
                return False

            encrypted_code = self._encrypt_password(
                code,
                salt=user.password_salt,
                iterations=user.hmac_iterations,
            )

            user_backup_code = (
                session.query(UserBackupCode)
                .filter_by(user_id=user.user_id, code=encrypted_code)
                .first()
            )

            if not user_backup_code:
                return False

            session.delete(user_backup_code)
            session.commit()

        return True

    def validate_otp_code(self, username: str, code: str) -> bool:
        otp = get_plugin('otp')
        assert otp

        with self._get_session() as session:
            user = self._get_user(session, username)
            if not user:
                return False

            otp_secret = self.get_otp_secret(username)
            if not otp_secret:
                return False

        return otp.verify_time_otp(otp=code, secret=otp_secret).output

    def disable_otp(self, username: str):
        with self._get_session(locked=True) as session:
            user = self._get_user(session, username)
            if not user:
                return False

            session.query(UserOtp).filter_by(user_id=user.user_id).delete()
            session.query(UserBackupCode).filter_by(user_id=user.user_id).delete()
            session.commit()
            return True

    def enable_otp(
        self,
        username: str,
        dry_run: bool = False,
        otp_secret: Optional[str] = None,
    ):
        with self._get_session() as session:
            user = self._get_user(session, username)
            if not user:
                raise InvalidCredentialsException()

            user_otp = session.query(UserOtp).filter_by(user_id=user.user_id).first()
            if user_otp:
                raise OtpRecordAlreadyExistsException()

            user_otp = self.create_otp_secret(
                username, otp_secret=otp_secret, dry_run=dry_run
            )

            backup_codes = (
                self.refresh_user_backup_codes(username) if not dry_run else []
            )

            return user_otp, backup_codes

    def generate_api_token(
        self,
        username: str,
        name: Optional[str] = None,
        expires_at: Optional[datetime.datetime] = None,
    ) -> str:
        """
        Create a random API token for a user.

        These are randomly generated tokens stored encrypted in the server's
        database. They are recommended for API usage over JWT tokens because:

            1. JWT tokens rely on the user's password and are invalidated if
               the user changes the password.
            2. JWT tokens are stateless and can't be revoked once generated -
               unless the user changes the password.
            3. They can end up exposing either the user's password, the
               server's keys, or both, if not handled properly.

        :param username: User name.
        :param name: Name of the token (default: ``<username>__<random-string>``).
        :param expires_at: Expiration datetime of the token.
        :return: The generated token as a string.
        :raises: :class:`platypush.exceptions.user.NoUserException` if the user
            does not exist.
        """
        user = self.get_user(username)
        if not user:
            raise NoUserException()

        token = (
            username
            + ':'
            + ''.join(
                random.choice(
                    "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789._"
                )
                for _ in range(32)
            )
        )

        name = name or f'{username}__' + ''.join(
            random.choice("ABCDEFGHIJKLMNOPQRSTUVWXYZ234567") for _ in range(8)
        )

        with self._get_session() as session:
            user_token = UserToken(
                user_id=user.user_id,
                name=name,
                token=self._encrypt_password(
                    token, user.password_salt, user.hmac_iterations
                ),
                created_at=utcnow(),
                expires_at=expires_at,
            )

            session.add(user_token)
            session.commit()

        return token

    @staticmethod
    def _user_by_token(session, token: str) -> Optional[User]:
        username = token.split(':')[0]
        return session.query(User).filter_by(username=username).first()

    def validate_api_token(self, token: str) -> User:
        """
        Validate an API token.

        :param token: Token to validate.
        :return: On success, it returns the user associated to the token.
        :raises: :class:`platypush.exceptions.user.InvalidTokenException` in
            case of invalid token.
        """
        with self._get_session() as session:
            user = self._user_by_token(session, token)
            if not user:
                raise InvalidTokenException()

            encrypted_token = self._encrypt_password(
                token, user.password_salt, user.hmac_iterations  # type: ignore
            )

            user_token = (
                session.query(UserToken)
                .filter(
                    and_(
                        UserToken.token == encrypted_token,
                        UserToken.user_id == user.user_id,
                        or_(
                            UserToken.expires_at.is_(None),
                            UserToken.expires_at >= utcnow(),
                        ),
                    )
                )
                .first()
            )

            if not user_token:
                raise InvalidTokenException()

            return user

    def delete_api_token(
        self,
        username: str,
        token: Optional[str] = None,
        token_id: Optional[int] = None,
    ):
        """
        Delete an API token.

        Either <token> or <token_id> must be provided.

        :param token: Token to delete.
        :param username: User name.
        :param token_id: Token ID.
        :return: True if the token was successfully deleted, False otherwise.
        """
        assert token or token_id, 'Either token or token_id must be provided'

        with self._get_session() as session:
            if token:
                user = self._user_by_token(session, token)
            else:
                user = self._get_user(session, username)

            assert user, 'No such user'

            if token_id:
                user_token = (
                    session.query(UserToken)
                    .filter_by(user_id=user.user_id, id=token_id)
                    .first()
                )
            else:
                encrypted_token = self._encrypt_password(
                    token, user.password_salt, user.hmac_iterations  # type: ignore
                )

                user_token = (
                    session.query(UserToken)
                    .filter_by(
                        token=encrypted_token,
                        user_id=user.user_id,
                    )
                    .first()
                )

            assert user_token, 'No such token'
            session.delete(user_token)
            session.commit()

    def get_api_tokens(self, username: str) -> List[UserToken]:
        """
        Get all the API tokens for a user.

        :param username: User name.
        :return: List of tokens.
        """
        with self._get_session() as session:
            user = self._get_user(session, username)
            if not user:
                return []

            return (
                session.query(UserToken)
                .filter(
                    and_(
                        UserToken.user_id == user.user_id,
                        or_(
                            UserToken.expires_at.is_(None),
                            UserToken.expires_at >= utcnow(),
                        ),
                    )
                )
                .order_by(UserToken.created_at.desc())
                .all()
            )


# vim:sw=4:ts=4:et:



import logging
import os
import threading
import time
from abc import ABC, abstractmethod
from typing import Any, Callable, Optional, Type

from platypush.message.event.media import (
    MediaDownloadCancelledEvent,
    MediaDownloadClearEvent,
    MediaDownloadCompletedEvent,
    MediaDownloadErrorEvent,
    MediaDownloadEvent,
    MediaDownloadPausedEvent,
    MediaDownloadProgressEvent,
    MediaDownloadResumedEvent,
    MediaDownloadStartedEvent,
)
from platypush.plugins.media._model import DownloadState
from platypush.plugins.media._resource import MediaResource


class MediaResourceDownloader(ABC):
    """
    Base media resource downloader class.
    """

    def __init__(self, media_plugin, *_, **__):
        from platypush.plugins.media import MediaPlugin

        self._media: MediaPlugin = media_plugin

    @abstractmethod
    def download(
        self, resource: MediaResource, path: Optional[str] = None, **_
    ) -> 'DownloadThread':
        pass

    def get_download_path(
        self,
        resource: MediaResource,
        *_,
        directory: Optional[str] = None,
        filename: Optional[str] = None,
        **__,
    ) -> str:
        directory = (
            os.path.expanduser(directory) if directory else self._media.download_dir
        )

        if not filename:
            filename = resource.filename or resource.url.split('/')[-1]

        return os.path.join(directory, filename)

    @abstractmethod
    def supports(self, resource_type: MediaResource) -> bool:
        return False

    def supports_only_audio(self) -> bool:
        return False


class DownloadThread(threading.Thread, ABC):
    """
    Thread that downloads a URL to a file.
    """

    _progress_update_interval = 1
    """ Throttle the progress updates to this interval, in seconds. """

    def __init__(
        self,
        path: str,
        url: str,
        post_event: Callable,
        size: Optional[int] = None,
        timeout: Optional[int] = 10,
        on_start: Callable[['DownloadThread'], None] = lambda _: None,
        on_close: Callable[['DownloadThread'], None] = lambda _: None,
        stop_event: Optional[threading.Event] = None,
    ):
        super().__init__(name=f'DownloadThread-{path}')
        self.path = path
        self.url = url
        self.size = size
        self.timeout = timeout
        self.state = DownloadState.IDLE
        self.progress = None
        self.started_at = None
        self.ended_at = None
        self._upstream_stop_event = stop_event or threading.Event()
        self._stop_event = threading.Event()
        self._post_event = post_event
        self._on_start = on_start
        self._on_close = on_close
        self._paused = threading.Event()
        self._downloading = threading.Event()
        self._last_progress_update_time = 0
        self.logger = logging.getLogger(__name__)

    def should_stop(self) -> bool:
        return self._stop_event.is_set() or self._upstream_stop_event.is_set()

    @abstractmethod
    def _run(self) -> bool:
        pass

    def pause(self):
        self.state = DownloadState.PAUSED
        self._paused.set()
        self._downloading.clear()
        self.post_event(MediaDownloadPausedEvent)

    def resume(self):
        self.state = DownloadState.DOWNLOADING
        self._paused.clear()
        self._downloading.set()
        self.post_event(MediaDownloadResumedEvent)

    def run(self):
        super().run()
        interrupted = False

        try:
            self.on_start()
            interrupted = not self._run()

            if interrupted:
                self.state = DownloadState.CANCELLED
            else:
                self.state = DownloadState.COMPLETED
        except Exception as e:
            self.state = DownloadState.ERROR
            self.post_event(MediaDownloadErrorEvent, error=str(e))
            self.logger.warning('Error while downloading URL: %s', e)
        finally:
            self.on_close()

    def stop(self):
        self.state = DownloadState.CANCELLED
        self._stop_event.set()
        self._downloading.clear()

    def on_start(self):
        self.state = DownloadState.STARTED
        self.started_at = time.time()
        self.post_event(MediaDownloadStartedEvent)
        self._on_start(self)

    def on_close(self):
        self.ended_at = time.time()
        if self.state == DownloadState.CANCELLED:
            self.post_event(MediaDownloadCancelledEvent)
        elif self.state == DownloadState.COMPLETED:
            self.post_event(MediaDownloadCompletedEvent)

        self._on_close(self)

    def clear(self):
        if self.state not in (DownloadState.COMPLETED, DownloadState.CANCELLED):
            self.logger.info(
                'Download thread for %s is still active, stopping', self.url
            )

            self.stop()
            self.join(timeout=10)

        self.post_event(MediaDownloadClearEvent)

    def post_event(self, event_type: Type[MediaDownloadEvent], **kwargs):
        kwargs = {
            'resource': self.url,
            'path': self.path,
            'state': self.state.value,
            'size': self.size,
            'timeout': self.timeout,
            'progress': self.progress,
            'started_at': self.started_at,
            'ended_at': self.ended_at,
            **kwargs,
        }

        self._post_event(event_type, **kwargs)

    def __setattr__(self, name: str, value: Optional[Any], /) -> None:
        if name == 'progress' and value is not None:
            if value < 0 or value > 100:
                self.logger.debug('Invalid progress value:%s', value)
                return

            prev_progress = getattr(self, 'progress', None)

            if prev_progress is None or (
                int(prev_progress) != int(value)
                and (
                    time.time() - self._last_progress_update_time
                    >= self._progress_update_interval
                )
            ):
                value = round(value, 2)
                self._last_progress_update_time = time.time()
                self.post_event(MediaDownloadProgressEvent, progress=value)

        super().__setattr__(name, value)


# vim:sw=4:ts=4:et:



from platypush.context import get_plugin
from platypush.plugins.media._search import MediaSearcher


# pylint: disable=too-few-public-methods
class TorrentMediaSearcher(MediaSearcher):
    """
    Media searcher for torrents.

    It needs at least one torrent plugin to be configured.
    """

    def search(self, query: str, *_, **__):
        self.logger.info('Searching torrents for "%s"', query)

        torrents = get_plugin(
            self.media_plugin.torrent_plugin if self.media_plugin else 'torrent'
        )

        if not torrents:
            raise RuntimeError('Torrent plugin not available/configured')

        return [
            torrent
            for torrent in torrents.search(
                query,
            ).output
            if torrent.get('is_media')
        ]

    def supports(self, type: str) -> bool:
        return type == 'torrent'


# vim:sw=4:ts=4:et:



import os
import threading
from typing import Collection, Optional

from platypush.context import get_bus
from platypush.plugins.media import PlayerState, MediaPlugin, MediaResource
from platypush.message.event.media import (
    MediaPlayEvent,
    MediaPlayRequestEvent,
    MediaPauseEvent,
    MediaStopEvent,
    MediaSeekEvent,
    MediaVolumeChangedEvent,
    MediaMuteChangedEvent,
    NewPlayingMediaEvent,
)

from platypush.plugins import action


class MediaVlcPlugin(MediaPlugin):
    """
    Plugin to control VLC instances.
    """

    def __init__(
        self,
        args: Optional[Collection[str]] = None,
        fullscreen: bool = False,
        volume: int = 100,
        **kwargs,
    ):
        """
        :param args: List of extra arguments to pass to the VLC executable (e.g.
            ``['--sub-language=en', '--snapshot-path=/mnt/snapshots']``)
        :param fullscreen: Set to True if you want media files to be opened in
            fullscreen by default (can be overridden by `.play()`) (default: False)
        :param volume: Default media volume (default: 100)
        """

        super().__init__(**kwargs)

        self._args = list(args or [])
        if '--play-and-exit' not in self._args:
            self._args.append('--play-and-exit')

        self._instance = None
        self._player = None
        self._latest_seek = None
        self._default_fullscreen = fullscreen
        self._default_volume = volume
        self._on_stop_callbacks = []
        self._monitor_thread: Optional[threading.Thread] = None
        self._on_stop_event = threading.Event()
        self._stop_lock = threading.RLock()

    @classmethod
    def _watched_event_types(cls):
        import vlc

        return [
            getattr(vlc.EventType, evt)
            for evt in [
                'MediaPlayerLengthChanged',
                'MediaPlayerMediaChanged',
                'MediaDurationChanged',
                'MediaPlayerMuted',
                'MediaPlayerUnmuted',
                'MediaPlayerOpening',
                'MediaPlayerPaused',
                'MediaPlayerPlaying',
                'MediaPlayerPositionChanged',
                'MediaPlayerStopped',
                'MediaPlayerTimeChanged',
                'MediaStateChanged',
                'MediaPlayerForward',
                'MediaPlayerBackward',
                'MediaPlayerEndReached',
                'MediaPlayerTitleChanged',
                'MediaPlayerAudioVolume',
            ]
            if hasattr(vlc.EventType, evt)
        ]

    def _init_vlc(self, resource: MediaResource, cache_streams: bool):
        if self._instance:
            self.logger.info('Another instance is running, waiting for it to terminate')
            self._on_stop_event.wait()

        for k, v in self._env.items():
            os.environ[k] = v

        self._monitor_thread = threading.Thread(target=self._player_monitor)
        self._monitor_thread.start()
        self._set_media(resource, cache_streams=cache_streams)
        assert self._player, 'Could not create a VLC player instance'

        for evt in self._watched_event_types():
            self._player.event_manager().event_attach(
                eventtype=evt, callback=self._event_callback()
            )

    def _set_media(
        self, resource: MediaResource, *_, cache_streams: bool = False, **__
    ):
        import vlc

        if not self._instance:
            self._instance = vlc.Instance(*self._args)

        assert self._instance, 'Could not create a VLC instance'
        if not self._player:
            self._player = self._instance.media_player_new()

        fd = resource.fd or resource.open(cache_streams=cache_streams)

        if not cache_streams and fd is not None:
            self._player.set_media(self._instance.media_new_fd(fd.fileno()))
        else:
            self._player.set_media(self._instance.media_new(resource.resource))

    def _player_monitor(self):
        self._on_stop_event.wait()
        self.logger.info('VLC stream terminated')
        self.quit()

    def _reset_state(self):
        self._latest_seek = None

        if self._latest_resource:
            self.logger.debug('Closing latest resource')
            self._latest_resource.close()
            self._latest_resource = None

    def _close_player(self):
        if self._player:
            self.logger.info('Releasing VLC player resource')
            self._player.stop()

            if self._player.get_media():
                self.logger.debug('Releasing VLC media resource')
                self._player.get_media().release()

            self.logger.debug('Releasing VLC player instance')
            self._player.release()
            self._player = None

        if self._instance:
            self.logger.info('Releasing VLC instance resource')
            self._instance.release()
            self._instance = None

    @staticmethod
    def _post_event(evt_type, **evt):
        bus = get_bus()
        bus.post(evt_type(player='local', plugin='media.vlc', **evt))

    @property
    def _title(self) -> Optional[str]:
        if not (self._player and self._player.get_media() and self._latest_resource):
            return None

        return (
            self._player.get_title()
            or self._latest_resource.title
            or self._latest_resource.filename
            or self._player.get_media().get_mrl()
            or None
        )

    def _event_callback(self):
        def callback(event):
            from vlc import EventType

            self.logger.debug('Received VLC event: %s', event.type)
            if event.type == EventType.MediaPlayerPlaying:  # type: ignore
                self._on_stop_event.clear()
                self._post_event(MediaPlayEvent, resource=self._get_current_resource())
            elif event.type == EventType.MediaPlayerPaused:  # type: ignore
                self._on_stop_event.clear()
                self._post_event(MediaPauseEvent)
            elif event.type in (
                EventType.MediaPlayerStopped,  # type: ignore
                EventType.MediaPlayerEndReached,  # type: ignore
            ):
                self._on_stop_event.set()
                self._post_event(MediaStopEvent)
                self._reset_state()
            elif self._player and (
                event.type
                in (
                    EventType.MediaPlayerTitleChanged,  # type: ignore
                    EventType.MediaPlayerMediaChanged,  # type: ignore
                )
            ):
                if event.type == EventType.MediaPlayerMediaChanged:  # type: ignore
                    self._post_event(NewPlayingMediaEvent, resource=self._title)
            elif event.type == EventType.MediaPlayerLengthChanged:  # type: ignore
                self._post_event(
                    NewPlayingMediaEvent, resource=self._get_current_resource()
                )
            elif self._player and event.type == EventType.MediaPlayerTimeChanged:  # type: ignore
                pos = float(self._player.get_time() / 1000)
                if self._latest_seek is None or abs(pos - self._latest_seek) > 5:
                    self._post_event(MediaSeekEvent, position=pos)
                self._latest_seek = pos
            elif self._player and event.type == EventType.MediaPlayerAudioVolume:  # type: ignore
                self._post_event(
                    MediaVolumeChangedEvent, volume=self._player.audio_get_volume()
                )
            elif event.type == EventType.MediaPlayerMuted:  # type: ignore
                self._post_event(MediaMuteChangedEvent, mute=True)
            elif event.type == EventType.MediaPlayerUnmuted:  # type: ignore
                self._post_event(MediaMuteChangedEvent, mute=False)
            elif event.type == EventType.MediaPlayerEncounteredError:  # type: ignore
                self.logger.error('VLC media player encountered an error')
                self._reset_state()

        return callback

    @action
    def play(
        self,
        resource: Optional[str] = None,
        subtitles: Optional[str] = None,
        fullscreen: Optional[bool] = None,
        volume: Optional[int] = None,
        cache_streams: Optional[bool] = None,
        metadata: Optional[dict] = None,
        **_,
    ):
        """
        Play a resource.

        :param resource: Resource to play - can be a local file or a remote URL
            (default: None == toggle play).
        :param subtitles: Path to optional subtitle file
        :param fullscreen: Set to explicitly enable/disable fullscreen (default:
            `fullscreen` configured value or False)
        :param volume: Set to explicitly set the playback volume (default:
            `volume` configured value or 100)
        :param cache_streams: Overrides the ``cache_streams`` configuration
            value.
        :param metadata: Optional metadata to pass to the resource.
        """

        if not resource:
            return self.pause()

        self._post_event(MediaPlayRequestEvent, resource=resource)
        cache_streams = (
            cache_streams if cache_streams is not None else self.cache_streams
        )
        media = self._get_resource(resource, metadata=metadata)
        self._latest_resource = media
        self.quit()
        self._init_vlc(media, cache_streams=cache_streams)
        if subtitles and self._player:
            if subtitles.startswith('file://'):
                subtitles = subtitles[len('file://') :]
            self._player.video_set_subtitle_file(subtitles)

        if self._player:
            self._player.play()

        if self.volume:
            self.set_volume(volume=self.volume)

        if fullscreen or self._default_fullscreen:
            self.set_fullscreen(True)

        if volume is not None or self._default_volume is not None:
            self.set_volume(volume if volume is not None else self._default_volume)

        return self.status()

    @action
    def pause(self, *_, **__):
        """Toggle the paused state"""
        assert self._player, 'No vlc instance is running'
        assert self._player.can_pause(), 'The specified media type cannot be paused'
        self._player.pause()
        return self.status()

    @action
    def quit(self, *_, **__):
        """Quit the player (same as `stop`)"""
        with self._stop_lock:
            if not self._player:
                self.logger.debug('No vlc instance is running')
                return self.status()

            self._reset_state()
            self._close_player()
            self._on_stop_event.wait(timeout=5)

        for cbk in self._on_stop_callbacks:
            cbk()

        return self.status()

    @action
    def stop(self, *_, **__):
        """Stop the application (same as `quit`)"""
        return self.quit()

    @action
    def voldown(self, *_, step: float = 10.0, **__):
        """Volume down by (default: 10)%"""
        assert self._player, 'No vlc instance is running'
        return self.set_volume(int(max(0, self._player.audio_get_volume() - step)))

    @action
    def volup(self, *_, step: float = 10.0, **__):
        """Volume up by (default: 10)%"""
        assert self._player, 'No vlc instance is running'
        return self.set_volume(int(min(100, self._player.audio_get_volume() + step)))

    @action
    def set_volume(self, volume: int):
        """
        Set the volume

        :param volume: Volume value between 0 and 100
        """
        assert self._player, 'No vlc instance is running'
        volume = max(0, min([100, volume]))
        self._player.audio_set_volume(volume)
        status: dict = self.status().output  # type: ignore
        status['volume'] = volume
        return status

    @action
    def seek(self, position: float, **__):
        """
        Seek backward/forward by the specified number of seconds

        :param position: Number of seconds relative to the current cursor
        """
        if not self._player:
            return None, 'No vlc instance is running'
        if not self._player.is_seekable():
            return None, 'The resource is not seekable'

        media = self._player.get_media()
        if not media:
            return None, 'No media loaded'

        pos = min(media.get_duration() / 1000, max(0, position))
        self._player.set_time(int(pos * 1000))
        return self.status()

    @action
    def back(self, *_, offset: float = 30.0, **__):
        """Back by (default: 30) seconds"""
        if not self._player:
            return None, 'No vlc instance is running'

        media = self._player.get_media()
        if not media:
            return None, 'No media loaded'

        pos = max(0, (self._player.get_time() / 1000) - offset)
        return self.seek(pos)

    @action
    def forward(self, *_, offset: float = 30.0, **__):
        """Forward by (default: 30) seconds"""
        if not self._player:
            return None, 'No vlc instance is running'

        media = self._player.get_media()
        if not media:
            return None, 'No media loaded'

        pos = min(
            media.get_duration() / 1000, (self._player.get_time() / 1000) + offset
        )
        return self.seek(pos)

    @action
    def toggle_subtitles(self, *_, **__):
        """Toggle the subtitles visibility"""
        assert self._player, 'No vlc instance is running'
        assert (
            self._player.video_get_spu_count() > 0
        ), 'The media file has no subtitles set'

        if self._player.video_get_spu() is None or self._player.video_get_spu() == -1:
            self._player.video_set_spu(0)
        else:
            self._player.video_set_spu(-1)

    @action
    def toggle_fullscreen(self):
        """Toggle the fullscreen mode"""
        assert self._player, 'No vlc instance is running'
        self._player.toggle_fullscreen()

    @action
    def set_fullscreen(self, fullscreen: bool = True):
        """Set fullscreen mode"""
        assert self._player, 'No vlc instance is running'
        self._player.set_fullscreen(fullscreen)

    @action
    def set_subtitles(self, filename: str, *_, **__):
        """Sets media subtitles from filename"""
        assert self._player, 'No vlc instance is running'
        if filename.startswith('file://'):
            filename = filename[len('file://') :]

        self._player.video_set_subtitle_file(filename)

    @action
    def remove_subtitles(self, *_, **__):
        """Removes (hides) the subtitles"""
        assert self._player, 'No vlc instance is running'
        self._player.video_set_spu(-1)

    @action
    def is_playing(self, *_, **__):
        """
        :returns: True if it's playing, False otherwise
        """
        if not self._player:
            return False
        return self._player.is_playing()

    @action
    def load(self, resource, *_, **args):
        """
        Load/queue a resource/video to the player
        """
        if not self._player:
            return self.play(resource, **args)

        media = self._get_resource(resource)
        self._reset_state()
        self._set_media(media)
        return self.status()

    @action
    def mute(self, *_, **__):
        """Toggle mute state"""
        assert self._player, 'No vlc instance is running'
        self._player.audio_toggle_mute()

    @action
    def set_position(self, position: float, **_):
        """
        Seek backward/forward to the specified absolute position (same as ``seek``)
        """
        return self.seek(position)

    @action
    def status(self):
        """
        Get the current player state.

        :returns: A dictionary containing the current state.

        Example::

            output = {
                "filename": "filename or stream URL",
                "state": "play"  # or "stop" or "pause"
            }

        """
        import vlc

        with self._stop_lock:
            if not (self._player and self._latest_resource):
                return {'state': PlayerState.STOP.value}

            status = self._latest_resource.to_dict()
            vlc_state = self._player.get_state()

            if vlc_state == vlc.State.Playing:  # type: ignore
                status['state'] = PlayerState.PLAY.value
            elif vlc_state == vlc.State.Paused:  # type: ignore
                status['state'] = PlayerState.PAUSE.value
            else:
                status['state'] = PlayerState.STOP.value

            status['position'] = (
                float(self._player.get_time() / 1000)
                if self._player.get_time() is not None
                else None
            )

            media = self._player.get_media()
            status['duration'] = status.get(
                'duration',
                (
                    media.get_duration() / 1000
                    if media and media.get_duration() is not None
                    else None
                ),
            )

            status['seekable'] = status['duration'] is not None
            status['fullscreen'] = self._player.get_fullscreen()
            status['mute'] = self._player.audio_get_mute()
            status['path'] = status['url']
            status['pause'] = status['state'] == PlayerState.PAUSE.value
            status['percent_pos'] = self._player.get_position() * 100
            status['filename'] = self._latest_resource.filename
            status['title'] = self._title
            status['volume'] = self._player.audio_get_volume()
            status['volume_max'] = 100
            return status

    def on_stop(self, callback):
        self._on_stop_callbacks.append(callback)

    def _get_current_resource(self):
        if not self._player or not self._player.get_media():
            return None

        if self._latest_resource:
            return self._latest_resource.url

        return self._player.get_media().get_mrl()


# vim:sw=4:ts=4:et:



from abc import ABC, abstractmethod
from logging import getLogger
from typing import Optional

from .. import MediaResource


# pylint: disable=too-few-public-methods
class MediaResourceParser(ABC):
    """
    Base class for media resource parsers.
    """

    def __init__(self, media_plugin, *_, **__):
        from platypush.plugins.media import MediaPlugin

        self._media: MediaPlugin = media_plugin
        self.logger = getLogger(self.__class__.__name__)

    @abstractmethod
    def parse(self, resource: str, *_, **__) -> Optional[MediaResource]:
        """
        Parses a media resource and returns a MediaResource object.
        """
        raise NotImplementedError


# vim:sw=4:ts=4:et:



import logging
from enum import Enum

from sqlalchemy import (
    Column,
    Enum as DbEnum,
    ForeignKey,
    Integer,
    JSON,
    String,
)

from platypush.common.db import is_defined

from . import Entity

logger = logging.getLogger(__name__)


class ProcedureType(Enum):
    PYTHON = 'python'
    CONFIG = 'config'
    DB = 'db'


if not is_defined('procedure'):

    class Procedure(Entity):
        """
        Models a procedure entity.
        """

        __tablename__ = 'procedure'

        id = Column(
            Integer, ForeignKey('entity.id', ondelete='CASCADE'), primary_key=True
        )
        args = Column(JSON, nullable=False, default=[])
        procedure_type = Column(
            DbEnum(
                *[m.value for m in ProcedureType.__members__.values()],
                name='procedure_type',
                create_constraint=True,
                validate_strings=True,
            ),
            nullable=False,
        )
        module = Column(String)
        source = Column(String)
        line = Column(Integer)
        actions = Column(JSON)

        __table_args__ = {'keep_existing': True}
        __mapper_args__ = {
            'polymorphic_identity': __tablename__,
        }



from marshmallow import fields

from ._base import SystemBaseSchema
from ._battery import BatterySchema
from ._cpu import CpuSchema
from ._disk import DiskSchema
from ._fan import FanSchema
from ._memory import MemoryStatsSchema, SwapStatsSchema
from ._network import NetworkInterfaceSchema
from ._temperature import TemperatureSchema


class SystemInfoSchema(SystemBaseSchema):
    """
    Schema for system info.
    """

    cpu = fields.Nested(CpuSchema)
    memory = fields.Nested(MemoryStatsSchema)
    swap = fields.Nested(SwapStatsSchema)
    disks = fields.List(fields.Nested(DiskSchema))
    network = fields.List(fields.Nested(NetworkInterfaceSchema))
    temperature = fields.List(fields.Nested(TemperatureSchema))
    fans = fields.List(fields.Nested(FanSchema))
    battery = fields.Nested(BatterySchema)



import base64
from functools import wraps
from typing import Optional, Union

from flask import request, redirect
from flask.wrappers import Response

from platypush.config import Config
from platypush.user import User, UserManager

from ..logger import logger
from .status import UserAuthStatus

user_manager = UserManager()


def get_arg(req, name: str) -> Optional[str]:
    # The Flask way
    if hasattr(req, 'args'):
        return req.args.get(name)

    # The Tornado way
    if hasattr(req, 'arguments'):
        arg = req.arguments.get(name)
        if arg:
            return arg[0].decode()

    return None


def get_cookie(req, name: str) -> Optional[str]:
    cookie = req.cookies.get(name)
    if not cookie:
        return None

    # The Flask way
    if isinstance(cookie, str):
        return cookie

    # The Tornado way
    return cookie.value


def authenticate_token(req) -> Optional[User]:
    global_token = Config.get('user.global_token')
    user_token = None

    if 'X-Token' in req.headers:
        user_token = req.headers['X-Token']
    elif 'Authorization' in req.headers and req.headers['Authorization'].startswith(
        'Bearer '
    ):
        user_token = req.headers['Authorization'][7:]
    else:
        user_token = get_arg(req, 'token')

    if not user_token:
        return None

    try:
        # Stantard API token authentication
        return user_manager.validate_api_token(user_token)
    except Exception as e:
        try:
            # Legacy JWT token authentication
            return user_manager.validate_jwt_token(user_token)
        except Exception as ee:
            logger().debug(
                'Invalid token. API token error: %s, JWT token error: %s', e, ee
            )

            # Legacy global token authentication.
            # The global token should be specified in the configuration file,
            # as a root parameter named `token`.
            if bool(global_token and user_token == global_token):
                return User(username='__token__', user_id=1)

            logger().info(e)


def authenticate_user_pass(req):
    # Flask populates request.authorization
    if hasattr(req, 'authorization'):
        if not req.authorization:
            return False

        username = req.authorization.username
        password = req.authorization.password

    # Otherwise, check the Authorization header
    elif 'Authorization' in req.headers and req.headers['Authorization'].startswith(
        'Basic '
    ):
        auth = req.headers['Authorization'][6:]
        try:
            auth = base64.b64decode(auth)
        except ValueError:
            pass

        username, password = auth.decode().split(':', maxsplit=1)
    else:
        return False

    return user_manager.authenticate_user(username, password)


def authenticate_session(req) -> Optional[User]:
    user = None

    # Check the X-Session-Token header
    user_session_token = req.headers.get('X-Session-Token')

    # Check the `session_token` query/body parameter
    if not user_session_token:
        user_session_token = get_arg(req, 'session_token')

    # Check the `session_token` cookie
    if not user_session_token:
        user_session_token = get_cookie(req, 'session_token')

    if user_session_token:
        user, _ = user_manager.authenticate_user_session(user_session_token)[:2]

    return user


def authenticate(
    redirect_page='',
    skip_auth_methods=None,
    json=False,
):
    """
    Authentication decorator for Flask routes.
    """

    def decorator(f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            auth_status = get_auth_status(
                request,
                skip_auth_methods=skip_auth_methods,
            )

            if auth_status == UserAuthStatus.OK:
                return f(*args, **kwargs)

            if json:
                return auth_status.to_response()

            if auth_status == UserAuthStatus.REGISTRATION_REQUIRED:
                return redirect(
                    f'/register?redirect={redirect_page or request.url}', 307
                )

            if auth_status == UserAuthStatus.INVALID_CREDENTIALS:
                return redirect(f'/login?redirect={redirect_page or request.url}', 307)

            return Response(
                'Authentication required',
                401,
                {'WWW-Authenticate': 'Basic realm="Login required"'},
            )

        return wrapper

    return decorator


# pylint: disable=too-many-return-statements
def get_current_user_or_auth_status(
    req, skip_auth_methods=None
) -> Union[User, UserAuthStatus]:
    """
    Returns the current user if authenticated, and the authentication status if
    ``with_status`` is True.
    """
    n_users = user_manager.get_user_count()
    skip_methods = skip_auth_methods or []

    # User/pass HTTP authentication
    http_auth_ok = True
    if n_users > 0 and 'http' not in skip_methods:
        response = authenticate_user_pass(req)
        if response:
            user = response[0] if isinstance(response, tuple) else response
            if user:
                return user

    # Token-based authentication
    token_auth_ok = True
    if 'token' not in skip_methods:
        user = authenticate_token(req)
        if user:
            return user

    # Session token based authentication
    session_auth_ok = True
    if n_users > 0 and 'session' not in skip_methods:
        user = authenticate_session(req)
        if user:
            return user

        return UserAuthStatus.INVALID_CREDENTIALS

    # At least a user should be created before accessing an authenticated resource
    if n_users == 0 and 'session' not in skip_methods:
        return UserAuthStatus.REGISTRATION_REQUIRED

    if (  # pylint: disable=too-many-boolean-expressions
        ('http' not in skip_methods and http_auth_ok)
        or ('token' not in skip_methods and token_auth_ok)
        or ('session' not in skip_methods and session_auth_ok)
    ):
        return UserAuthStatus.OK

    return UserAuthStatus.INVALID_CREDENTIALS


def get_auth_status(req, skip_auth_methods=None) -> UserAuthStatus:
    """
    Check against the available authentication methods (except those listed in
    ``skip_auth_methods``) if the user is properly authenticated.
    """
    ret = get_current_user_or_auth_status(req, skip_auth_methods=skip_auth_methods)
    return UserAuthStatus.OK if isinstance(ret, User) else ret


def current_user() -> Optional[User]:
    """
    Returns the current user if authenticated.
    """
    ret = get_current_user_or_auth_status(request)
    return ret if isinstance(ret, User) else None



import json
import os
import pathlib
import shutil
import stat
from functools import lru_cache
from multiprocessing import RLock
from typing import Any, Iterable, List, Dict, Optional, Set, Union

from platypush.plugins import Plugin, action
from platypush.utils import get_mime_type, is_binary

Bookmarks = Union[Iterable[str], Iterable[Union[str, Dict[str, Any]]], Dict[str, Any]]


class FilePlugin(Plugin):
    """
    A plugin for general-purpose file methods
    """

    def __init__(self, *args, bookmarks: Optional[Bookmarks] = None, **kwargs):
        """
        :param bookmarks: A list/dictionary of bookmarks. Bookmarks will be
            shown in the file browser UI home page for easier access.

            Possible formats:

              .. code-block:: yaml

                bookmarks:
                    - /path/to/directory1
                    - /path/to/directory2

              .. code-block:: yaml

                bookmarks:
                    Movies: /path/to/movies
                    Music: /path/to/music

              .. code-block:: yaml

                bookmarks:
                    - name: Movies
                      path: /path/to/movies
                      icon:
                          class: fa fa-film

              .. code-block:: yaml

                bookmarks:
                    Movies:
                        name: Movies
                        path: /path/to/movies
                        icon:
                            url: /path/to/icon.png

        """
        super().__init__(*args, **kwargs)
        self._mime_types_lock = RLock()
        self._bookmarks = self._parse_bookmarks(bookmarks)

    def _parse_bookmarks(self, bookmarks: Bookmarks) -> Dict[str, Dict[str, Any]]:
        ret = {}
        if isinstance(bookmarks, (list, tuple, set)):
            for bookmark in bookmarks:
                if isinstance(bookmark, str):
                    ret[bookmark] = {'name': bookmark, 'path': bookmark}
                else:
                    ret[bookmark['name']] = bookmark
        elif isinstance(bookmarks, dict):
            ret.update(bookmarks)

        for name, bookmark in ret.items():
            ret[name] = (
                {'name': name, 'path': os.path.abspath(os.path.expanduser(bookmark))}
                if isinstance(bookmark, str)
                else bookmark
            )

        return ret

    @classmethod
    def _get_path(cls, filename):
        return os.path.abspath(os.path.expanduser(filename))

    @classmethod
    def _to_string(cls, content):
        try:
            return json.dumps(content)
        except (ValueError, TypeError):
            return str(content)

    @action
    def read(self, file: str):
        """
        Read and return the content of a (text) file.

        Note that this method should only be used for small text files, as it
        reads the entire file in memory.

        If you need to read large/binary files, consider using the
        ``GET /file?path=<path>`` HTTP API endpoint instead.

        :param file: Path of the file.
        """
        with open(self._get_path(file), 'r') as f:
            return f.read()

    @action
    def write(self, file: str, content: str):
        """
        Writes content to a specified (text) file. Previous content will be truncated.

        :param file: Path of the file.
        :param content: Content to write.
        """

        if not isinstance(content, str):
            content = self._to_string(content)

        with open(self._get_path(file), 'w') as f:
            f.write(content)

    @action
    def append(self, file: str, content):
        """
        Append content to a specified (text) file.

        :param file: Path of the file.
        :param content: Content to write.
        """

        if not isinstance(content, str):
            content = self._to_string(content)

        with open(self._get_path(file), 'a') as f:
            f.write(content)

    @action
    def getsize(self, file):
        """
        Get the size of the specified file in bytes.

        :param file: File path.
        """
        return os.path.getsize(self._get_path(file))

    @action
    def mkdir(self, directory: str, exist_ok=True, parents=True, mode=0o755):
        """
        Create a directory.

        :param directory: Directory name/path.
        :param exist_ok: If set and the directory already exist the method will
            not return an error (default: True).
        :param parents: If set and any of the parent directories in the path don't
            exist they will be created (analogous to mkdir -p) (default: True).
        :param mode: Access mode (default: 0755).
        """
        pathlib.Path(self._get_path(directory)).mkdir(
            parents=parents, exist_ok=exist_ok, mode=mode
        )

    @action
    def rmdir(self, directory: str, recursive: bool = False):
        """
        Remove a directory. The directory must be empty.

        :param directory: Directory name/path.
        :param recursive: If set, the directory and all its contents will be
            removed recursively (default: False).
        """
        directory = self._get_path(directory)
        if not recursive:
            pathlib.Path(directory).rmdir()
        else:
            shutil.rmtree(directory)

    @action
    def touch(self, file: str, mode=0o644):
        """
        Create/touch a file.

        :param file: File name/path.
        :param mode: File permissions (default: 0644).
        """
        pathlib.Path(self._get_path(file)).touch(mode=mode)

    @action
    def chmod(self, file: str, mode):
        """
        Change the mode/permissions of a file.

        :param file: File name/path.
        :param mode: New file permissions.
        """
        pathlib.Path(self._get_path(file)).chmod(mode=mode)

    @action
    def home(self) -> str:
        """
        Returns the current user's home directory.
        """
        return str(pathlib.Path.home())

    @action
    def rename(self, file: str, name: str):
        """
        Rename/move a file.

        :param file: File to rename.
        :param name: New file name.
        """
        pathlib.Path(self._get_path(file)).rename(self._get_path(name))

    @action
    def copy(self, source: str, target: str):
        """
        Copy a file.

        :param source: Source file.
        :param target: Destination file.
        """
        shutil.copy(self._get_path(source), self._get_path(target))

    @action
    def move(self, source: str, target: str):
        """
        Move a file.

        :param source: Source file.
        :param target: Destination file.
        """
        shutil.move(self._get_path(source), self._get_path(target))

    @action
    def link(self, file: str, target: str, symbolic=True):
        """
        Create a link to a file.

        :param file: File to symlink.
        :param target: Symlink path.
        :param symbolic: If True, then the target link will be a symbolic link. Otherwise,
            it will be a hard link (default: symbolic).
        """
        path = pathlib.Path(self._get_path(file))
        target = self._get_path(target)

        if symbolic:
            path.symlink_to(target)
        else:
            path.hardlink_to(target)

    @action
    def unlink(self, file: str):
        """
        Remove a file or symbolic link.

        :param file: File/link to remove.
        """
        pathlib.Path(self._get_path(file)).unlink()

    @action
    def list(
        self,
        path: Optional[str] = None,
        sort: str = 'name',
        reverse: bool = False,
    ) -> List[Dict[str, Any]]:
        """
        List a file or all the files in a directory.

        :param path: File or directory (default: root directory).
        :param sort: Sort the files by ``name``, ``size``, ``last_modified``
            or ``created`` time (default: ``name``).
        :param reverse: If set, the files will be sorted in descending order
            according to the specified ``sort`` field (default: False).
        :return: List of files in the specified path, or absolute path of the
            specified path if ``path`` is a file and it exists. Each item will
            contain the fields ``type`` (``file`` or ``directory``) and
            ``path``.
        """
        path = self._get_path(path or '/')
        assert path and os.path.exists(path), f'No such file or directory: {path}'

        if not os.path.isdir(path):
            return [
                {
                    'type': 'file',
                    'path': path,
                    'name': os.path.basename(path),
                    **self._get_file_info(path),
                }
            ]

        return sorted(
            [
                {
                    'type': (
                        'directory' if os.path.isdir(os.path.join(path, f)) else 'file'
                    ),
                    'path': os.path.join(path, f),
                    'name': os.path.basename(f),
                    **self._get_file_info(os.path.join(path, f)),
                }
                for f in os.listdir(path)
            ],
            key=lambda f: (f.get('type'), (f.get(sort) or 0)),
            reverse=reverse,
        )

    @staticmethod
    def _get_file_info(file: str) -> Dict[str, Any]:
        ret: dict = {'path': file}

        try:
            ret['size'] = os.path.getsize(file)
        except Exception:
            ret['size'] = None

        try:
            ret['last_modified'] = os.path.getmtime(file)
        except Exception:
            ret['last_modified'] = None

        try:
            ret['created'] = os.path.getctime(file)
        except Exception:
            ret['created'] = None

        try:
            stat_info = os.stat(file)
            ret.update(
                {
                    'permissions': stat.filemode(stat_info.st_mode),
                    'owner': stat_info.st_uid,
                    'group': stat_info.st_gid,
                }
            )
        except Exception:
            ret.update({'permissions': None, 'owner': None, 'group': None})

        return ret

    @action
    def info(self, files: Iterable[str]) -> Dict[str, Dict[str, str]]:
        """
        Retrieve information about a list of files.

        :param files: List of files.
        :return: Dict containing the information about each file. Example:

          .. code-block:: json

            {
                "/path/to/file": {
                    "path": "/path/to/file",
                    "name": "file",
                    "size": 1234,
                    "type": "file",
                    "mime_type": "application/octet-stream",
                    "last_modified": "2021-01-01T00:00:00",
                    "permissions": "rw-r--r--",
                    "owner": "user",
                    "group": "group",
                }
            }

        """
        with self._mime_types_lock:
            ret = {}
            for file in files:
                file = self._get_path(file)
                if not os.path.exists(file):
                    self.logger.warning('File not found: %s', file)
                    continue

                ret[file] = {
                    **self._get_file_info(file),
                    'mime_type': get_mime_type(file),
                }

        return ret

    @action
    def get_mime_types(
        self,
        files: Iterable[str],
        types: Optional[Iterable[str]] = None,
    ) -> Dict[str, str]:
        """
        Given a list of files or URLs, get their MIME types, or filter them by
        MIME type.

        :param files: List of files or URLs.
        :param types: Filter of MIME types to apply. Partial matches are
            allowed - e.g. 'video' will match all video types. No filter means
            that all the input resources will be returned with their respective
            MIME types.
        :return: Dict containing the filtered resources and their MIME types.
        """
        filter_types = set()
        for t in types or []:
            filter_types.add(t)
            tokens = t.split('/')
            for token in tokens:
                filter_types.add(token)

        with self._mime_types_lock:
            return self._get_mime_types(files, filter_types)

    @action
    def is_binary(self, file: str) -> bool:
        """
        :file: File path.
        :return: True if the file is binary, False otherwise.
        """
        with open(self._get_path(file), 'rb') as f:
            return is_binary(f.read(1024))

    @action
    def get_user_home(self) -> str:
        """
        :return: The current user's home directory.
        """
        return str(pathlib.Path.home())

    @action
    def get_bookmarks(self) -> Dict[str, Dict[str, Any]]:
        """
        :return: List of bookmarks. Example:

            .. code-block:: json

                {
                    "directory1": {
                        "name": "directory1",
                        "path": "/path/to/directory1"
                        "icon": {
                            "class": "fa fa-folder"
                        }
                    },

                    "directory2": {
                        "name": "directory2",
                        "path": "/path/to/directory2"
                        "icon": {
                            "url": "/path/to/icon.png"
                        }
                    }
                }

        """
        return self._bookmarks

    def _get_mime_types(
        self, files: Iterable[str], filter_types: Set[str]
    ) -> Dict[str, str]:
        ret = {}
        for file in files:
            try:
                mime_type = self._get_mime_type(file)
            except Exception as e:
                self.logger.warning('Error while getting MIME type for %s: %s', file, e)
                continue

            mime_tokens = {mime_type, *mime_type.split('/')}
            if not filter_types or any(token in filter_types for token in mime_tokens):
                ret[file] = mime_type

        return ret

    @lru_cache(maxsize=1024)  # noqa
    def _get_mime_type(self, file: str) -> str:
        if file.startswith('file://'):
            file = file[len('file://') :]

        return get_mime_type(file) or 'application/octet-stream'


# vim:sw=4:ts=4:et:



from ._model import Cpu, CpuFrequency, CpuInfo, CpuStats, CpuTimes
from ._schemas import (
    CpuFrequencySchema,
    CpuInfoSchema,
    CpuSchema,
    CpuStatsSchema,
    CpuTimesSchema,
)


__all__ = [
    "Cpu",
    "CpuFrequency",
    "CpuFrequencySchema",
    "CpuInfo",
    "CpuInfoSchema",
    "CpuSchema",
    "CpuStats",
    "CpuStatsSchema",
    "CpuTimes",
    "CpuTimesSchema",
]



from collections import namedtuple
from enum import Enum

from flask import jsonify

from platypush.user import AuthenticationStatus

StatusValue = namedtuple('StatusValue', ['code', 'error', 'message'])


class UserAuthStatus(Enum):
    """
    Models the status of the authentication.
    """

    OK = StatusValue(200, AuthenticationStatus.OK, 'OK')
    INVALID_AUTH_TYPE = StatusValue(
        400, AuthenticationStatus.INVALID_AUTH_TYPE, 'Invalid authentication type'
    )
    INVALID_CREDENTIALS = StatusValue(
        401, AuthenticationStatus.INVALID_CREDENTIALS, 'Invalid credentials'
    )
    INVALID_JWT_TOKEN = StatusValue(
        401, AuthenticationStatus.INVALID_JWT_TOKEN, 'Invalid JWT token'
    )
    INVALID_OTP_CODE = StatusValue(
        401, AuthenticationStatus.INVALID_OTP_CODE, 'Invalid OTP code'
    )
    INVALID_METHOD = StatusValue(
        405, AuthenticationStatus.INVALID_METHOD, 'Invalid method'
    )
    MISSING_OTP_CODE = StatusValue(
        401, AuthenticationStatus.MISSING_OTP_CODE, 'Missing OTP code'
    )
    MISSING_PASSWORD = StatusValue(
        400, AuthenticationStatus.MISSING_PASSWORD, 'Missing password'
    )
    INVALID_SESSION = StatusValue(
        401, AuthenticationStatus.INVALID_CREDENTIALS, 'Invalid session'
    )
    INVALID_TOKEN = StatusValue(
        400, AuthenticationStatus.INVALID_JWT_TOKEN, 'Invalid token'
    )
    MISSING_USERNAME = StatusValue(
        400, AuthenticationStatus.MISSING_USERNAME, 'Missing username'
    )
    PASSWORD_MISMATCH = StatusValue(
        400, AuthenticationStatus.PASSWORD_MISMATCH, 'Password mismatch'
    )
    REGISTRATION_DISABLED = StatusValue(
        401, AuthenticationStatus.REGISTRATION_DISABLED, 'Registrations are disabled'
    )
    REGISTRATION_REQUIRED = StatusValue(
        412, AuthenticationStatus.REGISTRATION_REQUIRED, 'Please create a user first'
    )
    UNKNOWN_ERROR = StatusValue(
        500, AuthenticationStatus.UNKNOWN_ERROR, 'Unknown error'
    )

    def to_dict(self):
        return {
            'code': self.value[0],
            'error': self.value[1].name,
            'message': self.value[2],
        }

    def to_response(self):
        return jsonify(self.to_dict()), self.value[0]

    @staticmethod
    def by_status(status: AuthenticationStatus):
        for auth_status in UserAuthStatus:
            if auth_status.value[1] == status:
                return auth_status

        return None



import json
import re
from contextlib import contextmanager
from dataclasses import dataclass
from multiprocessing import RLock
from random import randint
from typing import (
    Callable,
    Collection,
    Dict,
    Generator,
    Iterable,
    List,
    Optional,
    Union,
)

import yaml
from sqlalchemy.orm import Session

from platypush.context import get_plugin
from platypush.entities.managers.procedures import ProcedureEntityManager
from platypush.entities.procedures import Procedure, ProcedureType
from platypush.message.event.entities import EntityDeleteEvent
from platypush.plugins import RunnablePlugin, action
from platypush.plugins.db import DbPlugin
from platypush.utils import run

from ._serialize import ProcedureEncoder


@dataclass
class _ProcedureWrapper:
    name: str
    obj: Union[dict, Callable]


class ProceduresPlugin(RunnablePlugin, ProcedureEntityManager):
    """
    Utility plugin to run and store procedures as native entities.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._status_lock = RLock()

    @action
    def exec(self, procedure: Union[str, dict], *args, **kwargs):
        """
        Execute a procedure.

        :param procedure: Procedure name or definition. If a string is passed,
            then the procedure will be looked up by name in the configured
            procedures. If a dictionary is passed, then it should be a valid
            procedure definition with at least the ``actions`` key.
        :param args: Optional arguments to be passed to the procedure.
        :param kwargs: Optional arguments to be passed to the procedure.
        """
        if isinstance(procedure, str):
            return run(f'procedure.{procedure}', *args, **kwargs)

        assert isinstance(procedure, dict), 'Invalid procedure definition'
        procedure_name = procedure.get(
            'name', f'procedure_{f"{randint(0, 1 << 32):08x}"}'
        )

        actions = procedure.get('actions')
        assert actions and isinstance(
            actions, (list, tuple, set)
        ), 'Procedure definition should have at least the "actions" key as a list of actions'

        try:
            # Create a temporary procedure definition and execute it
            self._all_procedures[procedure_name] = {
                'name': procedure_name,
                'type': ProcedureType.CONFIG.value,
                'actions': list(actions),
                'args': procedure.get('args', []),
                '_async': False,
            }

            kwargs = {
                **procedure.get('args', {}),
                **kwargs,
            }

            return self.exec(procedure_name, *args, **kwargs)
        finally:
            self._all_procedures.pop(procedure_name, None)

    def _convert_procedure(
        self, name: str, proc: Union[dict, Callable, Procedure]
    ) -> Procedure:
        if isinstance(proc, Procedure):
            return proc

        metadata = self._serialize_procedure(proc, name=name)
        return Procedure(
            id=name,
            name=name,
            plugin=self,
            procedure_type=metadata['type'],
            module=metadata.get('module'),
            source=metadata.get('source'),
            line=metadata.get('line'),
            args=metadata.get('args', []),
            actions=metadata.get('actions', []),
            meta=metadata.get('meta', {}),
        )

    @action
    def status(self, *_, publish: bool = True, **__):
        """
        :param publish: If set to True (default) then the
            :class:`platypush.message.event.entities.EntityUpdateEvent` events
            will be published to the bus with the current configured procedures.
            Usually this should be set to True, unless you're calling this method
            from a context where you first want to retrieve the procedures and
            then immediately modify them. In such cases, the published events may
            result in race conditions on the entities engine.
        :return: The serialized configured procedures. Format:

            .. code-block:: json

                {
                    "procedure_name": {
                        "type": "python",
                        "module": "module_name",
                        "source": "/path/to/source.py",
                        "line": 42,
                        "args": ["arg1", "arg2"]
                    }
                }

        """
        with self._status_lock:
            self._sync_db_procedures()
            if publish:
                self.publish_entities(self._get_wrapped_procedures())

            return self._get_serialized_procedures()

    @action
    def save(
        self,
        name: str,
        actions: Iterable[dict],
        args: Optional[Iterable[str]] = None,
        old_name: Optional[str] = None,
        meta: Optional[dict] = None,
        **_,
    ):
        """
        Save a procedure.

        :param name: Name of the procedure.
        :param actions: Definition of the actions to be executed. Format:

              .. code-block:: json

                [
                    {
                        "action": "logger.info",
                        "args": {
                            "msg": "Hello, world!"
                        }
                    }
                ]

        :param args: Optional list of arguments to be passed to the procedure,
            as a list of strings with the argument names.
        :param old_name: Optional old name of the procedure if it's being
            renamed.
        :param meta: Optional metadata to be stored with the procedure. Example:

            .. code-block:: json

                {
                    "icon": {
                        "class": "fas fa-cogs",
                        "color": "#00ff00"
                    }
                }

        """
        assert name, 'Procedure name cannot be empty'
        assert actions, 'Procedure actions cannot be empty'

        args = args or []
        proc_def = self._all_procedures.get(name, {})
        proc_args = {
            'name': name,
            'type': ProcedureType.DB.value,
            'actions': actions,
            'args': args,
            'meta': (
                meta or (proc_def.get('meta', {}) if isinstance(proc_def, dict) else {})
            ),
        }

        def _on_entity_saved(*_, **__):
            self._all_procedures[name] = proc_args

        with self._status_lock:
            with self._db_session() as session:
                if old_name and old_name != name:
                    try:
                        self._delete(old_name, session=session)
                    except AssertionError as e:
                        self.logger.warning(
                            'Error while deleting old procedure: name=%s: %s',
                            old_name,
                            e,
                        )

            self.publish_entities(
                [_ProcedureWrapper(name=name, obj=proc_args)],
                callback=_on_entity_saved,
            )

        return self.status(publish=False)

    @action
    def delete(self, name: str):
        """
        Delete a procedure by name.

        Note that this is only possible for procedures that are stored on the
        database. Procedures that are loaded from Python scripts or
        configuration files should be removed from the source file.

        :param name: Name of the procedure to be deleted.
        """
        with self._db_session() as session:
            self._delete(name, session=session)

        self.status()

    @action
    def to_yaml(self, procedure: Union[str, dict]) -> str:
        """
        Serialize a procedure to YAML.

        This method is useful to export a procedure to a file.
        Note that it only works with either YAML-based procedures or
        database-stored procedures: Python procedures can't be converted to
        YAML.

        :param procedure: Procedure name or definition. If a string is passed,
            then the procedure will be looked up by name in the configured
            procedures. If a dictionary is passed, then it should be a valid
            procedure definition with at least the ``actions`` and ``name``
            keys.
        :return: The serialized procedure in YAML format.
        """
        if isinstance(procedure, str):
            proc = self._all_procedures.get(procedure)
            assert proc, f'Procedure {proc} not found'
        elif isinstance(procedure, dict):
            name = self._normalize_name(procedure.get('name'))
            assert name, 'Procedure name cannot be empty'

            actions = procedure.get('actions', [])
            assert actions and isinstance(
                actions, (list, tuple, set)
            ), 'Procedure definition should have at least the "actions" key as a list of actions'

            args = [self._normalize_name(arg) for arg in procedure.get('args', [])]
            proc = {
                f'procedure.{name}'
                + (f'({", ".join(args)})' if args else ''): [
                    self._serialize_action(action) for action in actions
                ]
            }
        else:
            raise AssertionError(
                f'Invalid procedure definition with type {type(procedure)}'
            )

        return yaml.safe_dump(proc, default_flow_style=False, indent=2)

    @staticmethod
    def _normalize_name(name: Optional[str]) -> str:
        return re.sub(r'[^\w.]+', '_', (name or '').strip(' .'))

    @classmethod
    def _serialize_action(cls, data: Union[Iterable, Dict]) -> Union[Dict, List, str]:
        if isinstance(data, dict):
            name = data.get('action', data.get('name'))
            if name:
                return {
                    'action': name,
                    **({'args': data['args']} if data.get('args') else {}),
                }

            return {
                k: (
                    cls._serialize_action(v)
                    if isinstance(v, (dict, list, tuple))
                    else v
                )
                for k, v in data.items()
            }
        elif isinstance(data, str):
            return data
        else:
            return [cls._serialize_action(item) for item in data if item is not None]

    @contextmanager
    def _db_session(self) -> Generator[Session, None, None]:
        db: Optional[DbPlugin] = get_plugin(DbPlugin)
        assert db, 'No database plugin configured'
        with db.get_session(locked=True) as session:
            assert isinstance(session, Session)
            yield session

            if session.is_active:
                session.commit()
            else:
                session.rollback()

    def _delete(self, name: str, session: Session):
        assert name, 'Procedure name cannot be empty'
        proc_row: Procedure = (
            session.query(Procedure).filter(Procedure.name == name).first()
        )

        assert proc_row, f'Procedure {name} not found in the database'
        assert proc_row.procedure_type == ProcedureType.DB.value, (  # type: ignore[attr-defined]
            f'Procedure {name} is not stored in the database, '
            f'it should be removed from the source file'
        )

        session.delete(proc_row)
        self._all_procedures.pop(name, None)
        self._bus.post(EntityDeleteEvent(plugin=self, entity=proc_row))

    def transform_entities(
        self, entities: Collection[_ProcedureWrapper], **_
    ) -> Collection[Procedure]:
        return [
            self._convert_procedure(
                name=proc.name,
                proc=proc if isinstance(proc, Procedure) else proc.obj,
            )
            for proc in entities
        ]

    def _get_wrapped_procedures(self) -> Collection[_ProcedureWrapper]:
        return [
            _ProcedureWrapper(name=name, obj=proc)
            for name, proc in self._all_procedures.items()
        ]

    def _sync_db_procedures(self):
        with self._status_lock:
            cur_proc_names = set(self._all_procedures.keys())
            with self._db_session() as session:
                saved_procs = {
                    str(proc.name): proc for proc in session.query(Procedure).all()
                }

                procs_to_remove = [
                    proc
                    for name, proc in saved_procs.items()
                    if name not in cur_proc_names
                    and proc.procedure_type != ProcedureType.DB.value  # type: ignore[attr-defined]
                ]

                for proc in procs_to_remove:
                    self.logger.info(
                        'Removing stale procedure record for %s', proc.name
                    )
                    session.delete(proc)

                procs_to_add = [
                    proc
                    for proc in saved_procs.values()
                    if proc.procedure_type == ProcedureType.DB.value  # type: ignore[attr-defined]
                ]

                for proc in procs_to_add:
                    self._all_procedures[str(proc.name)] = {
                        'type': proc.procedure_type,
                        'name': proc.name,
                        'args': proc.args,
                        'actions': proc.actions,
                        'meta': proc.meta,
                    }

    @staticmethod
    def _serialize_procedure(
        proc: Union[dict, Callable], name: Optional[str] = None
    ) -> dict:
        ret = json.loads(json.dumps(proc, cls=ProcedureEncoder))
        if name:
            ret['name'] = name

        return ret

    def _get_serialized_procedures(self) -> dict:
        return {
            name: self._serialize_procedure(proc, name=name)
            for name, proc in self._all_procedures.items()
        }

    def main(self, *_, **__):
        self._sync_db_procedures()

        while not self.should_stop():
            self.publish_entities(self._get_wrapped_procedures())
            self.wait_stop()
